"""
==========================
Direct Bundle Registration
==========================

This example explains how you can register two bundles from two different
subjects directly in the space of streamlines :footcite:p:`Garyfallidis2014b`,
:footcite:p:`Garyfallidis2015`.

To show the concept we will use two pre-saved cingulum bundles. The algorithm
used here is called Streamline-based Linear Registration (SLR)
:footcite:p:`Garyfallidis2015`.
"""

from time import sleep

from dipy.align.streamlinear import StreamlineLinearRegistration
from dipy.data import two_cingulum_bundles
from dipy.tracking.streamline import set_number_of_points
from dipy.viz import actor, window

###############################################################################
# Let's download and load the two bundles.

cb_subj1, cb_subj2 = two_cingulum_bundles()

###############################################################################
# An important step before running the registration is to resample the
# streamlines so that they both have the same number of points per streamline.
# Here we will use 20 points. This step is not optional. Inputting streamlines
# with a different number of points will break the theoretical advantages of
# using the SLR as explained in :footcite:p:`Garyfallidis2015`.

cb_subj1 = set_number_of_points(cb_subj1, nb_points=20)
cb_subj2 = set_number_of_points(cb_subj2, nb_points=20)

###############################################################################
# Let's say now that we want to move the ``cb_subj2`` (moving) so that it can
# be aligned with ``cb_subj1`` (static). Here is how this is done.

srr = StreamlineLinearRegistration()

srm = srr.optimize(static=cb_subj1, moving=cb_subj2)

###############################################################################
# After the optimization is finished we can apply the transformation to
# ``cb_subj2``.

cb_subj2_aligned = srm.transform(cb_subj2)


def show_both_bundles(bundles, colors=None, show=True, fname=None):
    scene = window.Scene()
    scene.SetBackground(1.0, 1, 1)
    for i, bundle in enumerate(bundles):
        color = colors[i]
        lines_actor = actor.streamtube(bundle, colors=color, linewidth=0.3)
        lines_actor.RotateX(-90)
        lines_actor.RotateZ(90)
        scene.add(lines_actor)
    if show:
        window.show(scene)
    if fname is not None:
        sleep(1)
        window.record(scene=scene, n_frames=1, out_path=fname, size=(900, 900))


show_both_bundles(
    [cb_subj1, cb_subj2],
    colors=[window.colors.orange, window.colors.red],
    show=False,
    fname="before_registration.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Before bundle registration.

show_both_bundles(
    [cb_subj1, cb_subj2_aligned],
    colors=[window.colors.orange, window.colors.red],
    show=False,
    fname="after_registration.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# After bundle registration.
#
#
#
# As you can see the two cingulum bundles are well aligned although they
# contain many streamlines of different lengths and shapes.
#
# Streamline-based Linear Registration (SLR) is a method which given two sets
# of streamlines (fixed and moving) and a streamline-based cost function, will
# minimize the cost function and transform the moving set of streamlines
# (target) to the fixed (reference), so that they maximally overlap under the
# condition that the transform stays linear.
#
# We denote a single streamline with s and a set of streamlines with S.
# A streamline s is an ordered sequence of line segments connecting 3D vector
# points $\mathbf{x}_{k} \in \mathbb{R}^{3}$ with $k \in[1, K]$ where K is the
# total number of points of streamline s. Given two bundles(two sets of
# streamlines), we denote $S_{a}=\left\{s_{1}^{a}, \ldots, S_{A}^{a}\right\}$
# and $S_{b}=\left\{s_{1}^{b}, \ldots, s_{B}^{b}\right\}$, where A and B are
# the total numbers of streamlines in each set respectively. We want to
# minimize a cost function so that we can align the two sets together. For
# this purpose, we introduce a new cost function, the Bundle-based Minimum
# Distance (BMD), which is defined as:
#
# .. math::
#
#     \operatorname{BMD}\left(S_{a}, S_{b}\right)=\frac{1}{4}\left(\frac{1}{A}
#         \sum_{i=1}^{A} \min _{j} D(i, j)+\frac{1}{B} \sum_{j=1}^{B} \\
#         \min _{i} D(i, j)\right)^{2}
#
#
# where D is the rectangular matrix given by all pairwise Minimum average
# Direct-Flip (MDF) streamline distances (Garyfallidis et al., 2012).
# Therefore, every element of matrix D is equal to
# $D_{i j}=M D F\left(s^{a}{ }_{i}, s^{b}{ }_{j}\right)$.
#
# Notice, how in Eq. (1), the most similar streamlines from one streamline set
# to the other are weighted more by averaging the minimum values of the rows
# and columns of matrix D. This makes our method robust to fanning streamlines
# near endpoints of bundles and spurious streamlines if any in the bundle. The
# MDF is a symmetric distance between two individual streamlines. It was
# primarily used for clustering (Garyfallidis et al., 2010; Visser et al.,
# 2011) and tractography simplification (see Garyfallidis et al., 2012). This
# distance can be applied only when both streamlines have the same number of
# points. Therefore we assume from now on that an initial interpolation of
# streamlines has been applied, so that all streamlines have the same number
# of points K, and all segments of each streamline have equal length. The
# length of each segment is equal to the length of the streamline divided by
# the number of segments $(K-1)$. This is achieved by a simple linear
# interpolation with the starting and ending points of the streamlines intact.
# When K is small, the interpolation provides a rough representation of the
# streamline, but as K becomes larger and larger the shape of the interpolated
# streamline becomes identical with the shape of the initial streamline.
# Under this assumption, the MDF for two streamlines $S_{a}$ and $S_{b}$ is
# defined as:
#
#
# .. math::
#
#     \operatorname{MDF}\left(s_{i}^{a}, s_{j}^{b}\right)=\min \\
#         \left(d_{\text {direct }}\left(s_{i}^{a}, s_{j}^{b}\right), \\
#         d_{\text {flipped }}\left(s_{i}^{a}, s_{j}^{b}\right)\right)
#
#
# where $d_{\text {direct }}$ is the direct distance which is defined as:
#
# .. math::
#
#     d_{\text {direct }}\left(s_{i}^{a}, s_{j}^{b}\right)=\frac{1}{K} \\
#         \sum_{k=1}^{K}\left\|\mathbf{x}_{k}^{a}-\mathbf{x}_{k}^{b}\right\|_{2}
#
# where $x_{k}^{a}$ is the k-th point of streamline $S_{i}^{a}$ and $x_{k}^{b}$
# is the k-th point of streamline $S_{j}^{b}$. $d_{\text {flipped }}$ is the
# one of the streamlines flipped and it is defined as:
#
# .. math::
#
#     d_{\text {flipped }}\left(s_{i}^{a}, s_{j}^{b}\right)=\frac{1}{K} \\
#         \sum_{k=1}^{K}\left\|\mathbf{x}_{k}^{a}-\mathbf{x}_{K-k+1}^{b}\\
#         \right\|_{2}
#
# and K is the total number of points in $x^{a}$ and $x^{b}$.
# The MDF has two very useful properties. First, it takes into consideration
# that streamlines have no preferred orientation. Second, it is a
# mathematically sound metric distance in the space of streamlines as proved
# in Garyfallidis et al. (2012). This means that the MDF is nonnegative, 0
# only when both streamlines are identical, symmetric and it satisfies the
# triangle inequality. Now that we have defined our cost function in Eq. (1)
# we can formulate the following optimization problem. Given a fixed bundle S
# and a moving bundle M we would like to find the vector of parameters t
# which transforms M to S using a linear transformation T so that BMD is
# minimum:
#
# .. math::
#
#     \operatorname{SLR}(S, M)=\\underset{\mathbf{t}}{\operatorname{argmin}} \\
#         \operatorname{BMD}(S, T(M, \mathbf{t}))
#
#
# Here, $\mathbf{t}$ is a vector in $\mathbb{R}^{n}$ holding the parameters of
# the linear transform where n = 12 for affine or n = 6 for rigid registration.
# From this vector we can then compose the transformation matrix which is
# applied to all the points of bundle M.
#
#
# References
# ----------
#
# .. footbibliography::



""" Detect common dtype across object array """

from functools import reduce

cimport numpy as cnp
cimport cython

import numpy as np

cdef enum:
    SCALAR, ARRAY

SCALAR_TYPES = np.ScalarType


@cython.boundscheck(False)
@cython.wraparound(False)
def quick_squash(obj_arr, mask=None, fill=0):
    """Try and make a standard array from an object array

    This function takes an object array and attempts to convert it to a more
    useful dtype. If array can be converted to a better dtype, Nones are
    replaced by `fill`. To make the behaviour of this function more clear, here
    are the most common cases:

    1.  `obj_arr` is an array of scalars of type `T`. Returns an array like
        `obj_arr.astype(T)`
    2.  `obj_arr` is an array of arrays. All items in `obj_arr` have the same
        shape ``S``. Returns an array with shape ``obj_arr.shape + S``
    3.  `obj_arr` is an array of arrays of different shapes. Returns `obj_arr`.
    4.  Items in `obj_arr` are not ndarrays or scalars. Returns `obj_arr`.

    Parameters
    ----------
    obj_arr : array, dtype=object
        The array to be converted.
    mask : array, dtype=bool, optional
       mask is nonzero where `obj_arr` has Nones.
    fill : number, optional
        Nones are replaced by `fill`.

    Returns
    -------
    result : array

    Examples
    --------
    >>> arr = np.empty(3, dtype=object)
    >>> arr.fill(2)
    >>> quick_squash(arr)
    array([2, 2, 2])
    >>> arr[0] = None
    >>> quick_squash(arr)
    array([0, 2, 2])
    >>> arr.fill(np.ones(2))
    >>> r = quick_squash(arr)
    >>> r.shape
    (3, 2)
    >>> r.dtype
    dtype('float64')
    """
    cdef:
        cnp.npy_intp i, j, N, dtypes_i
        object [:] flat_obj
        char [:] flat_mask
        cnp.dtype [:] dtypes
        int have_mask = not mask is None
        int search_for
        cnp.ndarray result
        cnp.dtype dtype, last_dtype
        object common_shape
    if have_mask:
        flat_mask = np.array(mask.reshape(-1), dtype=np.int8)
    N = obj_arr.size
    dtypes = np.empty((N,), dtype=object)
    flat_obj = obj_arr.reshape((-1))
    # Find first valid value
    for i in range(N):
        e = flat_obj[i]
        if ((have_mask and flat_mask[i] == 0) or
            (not have_mask and e is None)):
            continue
        t = type(e)
        if issubclass(t, np.generic) or t in SCALAR_TYPES:
            search_for = SCALAR
            common_shape = ()
            dtype = np.dtype(t)
            break
        elif t == cnp.ndarray:
            search_for = ARRAY
            common_shape = e.shape
            dtype = e.dtype
            break
        else: # something other than scalar or array
            return obj_arr
    else: # Nothing outside mask / all None
        return obj_arr
    # Check rest of values to confirm common type / shape, and collect dtypes
    last_dtype = dtype
    dtypes[0] = dtype
    dtypes_i = 1
    for j in range(i+1, N):
        e = flat_obj[j]
        if ((have_mask and flat_mask[j] == 0) or
            (not have_mask and e is None)):
            continue
        t = type(e)
        if search_for == SCALAR:
            if not issubclass(t, np.generic) and not t in SCALAR_TYPES:
                return obj_arr
            dtype = np.dtype(t)
        else: # search_for == ARRAY:
            if not t == cnp.ndarray:
                return obj_arr
            if not e.shape == common_shape:
                return obj_arr
            dtype = e.dtype
        if dtype != last_dtype:
            last_dtype = dtype
            dtypes[dtypes_i] = dtype
            dtypes_i += 1
    # Find common dtype
    unique_dtypes = set(dtypes[:dtypes_i])
    tiny_arrs = [np.zeros((1,), dtype=dt) for dt in unique_dtypes]
    dtype = reduce(np.add, tiny_arrs).dtype
    # Create and fill output array
    result = np.empty((N,) + common_shape, dtype=dtype)
    for i in range(N):
        e = flat_obj[i]
        if ((have_mask and flat_mask[i] == 0) or
            (not have_mask and e is None)):
            result[i] = fill
        else:
            result[i] = e
    return result.reshape(obj_arr.shape + common_shape)



"""
Decorators for dipy tests
"""

from functools import wraps
import inspect
from inspect import Parameter, signature
import os
import platform
import re
import warnings

import numpy as np
from packaging import version

import dipy

SKIP_RE = re.compile(r"(\s*>>>.*?)(\s*)#\s*skip\s+if\s+(.*)$")


def doctest_skip_parser(func):
    """Decorator replaces custom skip test markup in doctests.

    Say a function has a docstring::

        >>> something # skip if not HAVE_AMODULE
        >>> something + else
        >>> something # skip if HAVE_BMODULE

    This decorator will evaluate the expression after ``skip if``.  If this
    evaluates to True, then the comment is replaced by ``# doctest: +SKIP``.
    If False, then the comment is just removed. The expression is evaluated in
    the ``globals`` scope of `func`.

    For example, if the module global ``HAVE_AMODULE`` is False, and module
    global ``HAVE_BMODULE`` is False, the returned function will have
    docstring::
        >>> something # doctest: +SKIP
        >>> something + else
        >>> something

    """
    lines = func.__doc__.split("\n")
    new_lines = []
    for line in lines:
        match = SKIP_RE.match(line)
        if match is None:
            new_lines.append(line)
            continue
        code, space, expr = match.groups()
        if eval(expr, func.__globals__):
            code = code + space + "# doctest: +SKIP"
        new_lines.append(code)
    func.__doc__ = "\n".join(new_lines)
    return func


###
# In some cases (e.g., on Travis), we want to use a virtual frame-buffer for
# testing. The following decorator runs the tests under xvfb (mediated by
# xvfbwrapper) conditioned on an environment variable (that we set in
# .travis.yml for these cases):
use_xvfb = os.environ.get("TEST_WITH_XVFB", False)
is_windows = platform.system().lower() == "windows"
is_macOS = platform.system().lower() == "darwin"
is_linux = platform.system().lower() == "linux"


def xvfb_it(my_test):
    """Run a test with xvfbwrapper."""
    # When we use verbose testing we want the name:
    fname = my_test.__name__

    def test_with_xvfb(*args, **kwargs):
        if use_xvfb:
            from xvfbwrapper import Xvfb

            display = Xvfb(width=1920, height=1080)
            display.start()
        my_test(*args, **kwargs)
        if use_xvfb:
            display.stop()

    # Plant it back in and return the new function:
    test_with_xvfb.__name__ = fname
    return test_with_xvfb if not is_windows else my_test


def set_random_number_generator(seed_v=1234):
    """Decorator to use a fixed value for the random generator seed.

    This will make the tests that use random functions reproducible.

    """

    def _set_random_number_generator(func):
        def _set_random_number_generator_wrapper(pytestconfig, *args, **kwargs):
            rng = np.random.default_rng(seed_v)
            kwargs["rng"] = rng
            signature = inspect.signature(func)
            if pytestconfig and "pytestconfig" in signature.parameters.keys():
                output = func(pytestconfig, *args, **kwargs)
            else:
                output = func(*args, **kwargs)
            return output

        return _set_random_number_generator_wrapper

    return _set_random_number_generator


def warning_for_keywords(from_version="1.10.0", until_version="2.0.0"):
    """
    Decorator to warn about keyword arguments passed as positional arguments
    and handle version-based deprecation of functions.

    Parameters
    ----------
    from_version : str, optional
        The version from which the warning should start.
    until_version : str, optional
        The version until which the warning is applicable.

    Returns
    -------
    function
        The wrapped function that will issue a warning based
        on the version.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            current_version = dipy.__version__

            parsed_version = version.parse(current_version)
            current_version = parsed_version.base_version

            def convert_positional_to_keyword(func, args, kwargs):
                """
                Converts excess positional arguments to keyword arguments.

                Parameters
                ----------
                func : function
                    The original function to be called.
                args : tuple
                    The positional arguments passed to the function.
                kwargs : dict
                    The keyword arguments passed to the function.

                Returns
                -------
                result
                    The result of the function call with corrected arguments.
                """
                sig = signature(func)
                params = sig.parameters
                max_positional_args = sum(
                    1
                    for param in params.values()
                    if param.kind
                    in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)
                )

                if len(args) > max_positional_args:
                    # Split positional arguments into valid positional and
                    # those needing conversion
                    positional_args = list(args[:max_positional_args])
                    corrected_kwargs = dict(kwargs)
                    positionally_passed_kwonly_args = []
                    for param, arg in zip(
                        list(params.values())[max_positional_args:],
                        args[max_positional_args:],
                    ):
                        corrected_kwargs[param.name] = arg
                        positionally_passed_kwonly_args.append(param.name)

                    if positionally_passed_kwonly_args and version.parse(
                        from_version
                    ) <= version.parse(current_version) <= version.parse(until_version):
                        warnings.warn(
                            f"Pass {positionally_passed_kwonly_args} as keyword args. "
                            f"From version {until_version} passing these as positional "
                            f"arguments will result in an error. ",
                            UserWarning,
                            stacklevel=3,
                        )

                    return func(*positional_args, **corrected_kwargs)

                return func(*args, **kwargs)

            # Check if the current version is within the warning range
            if (
                version.parse(from_version)
                <= version.parse(current_version)
                <= version.parse(until_version)
            ):
                # Convert positional to keyword arguments and issue a warning
                return convert_positional_to_keyword(func, args, kwargs)

            # If the version is greater than the until_version,
            # pass the arguments as they are
            elif version.parse(current_version) > version.parse(until_version):
                return func(*args, **kwargs)

            # Convert positional to keyword arguments if
            # current version is less than from_version without warning
            elif version.parse(current_version) < version.parse(from_version):
                return convert_positional_to_keyword(func, args, kwargs)

            # Default case: call the function with the original arguments
            return func(*args, **kwargs)

        return wrapper

    return decorator



from numpy.testing import assert_array_almost_equal

from dipy.core.subdivide_octahedron import create_unit_sphere


def test_create_unit_sphere():
    sphere = create_unit_sphere(recursion_level=7)
    v, _, _ = sphere.vertices, sphere.edges, sphere.faces
    assert_array_almost_equal((v * v).sum(1), 1)


def create_half_unit_sphere():
    sphere = create_half_unit_sphere(7)
    v, _, _ = sphere.vertices, sphere.edges, sphere.faces
    assert_array_almost_equal((v * v).sum(1), 1)



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from dipy.core.gradients import gradient_table
from dipy.core.sphere_stats import angular_similarity
from dipy.core.subdivide_octahedron import create_unit_sphere
from dipy.data import default_sphere, dsi_voxels, get_fnames, get_sphere
from dipy.direction.peaks import peak_directions
from dipy.reconst.gqi import GeneralizedQSamplingModel
from dipy.reconst.odf import gfa
from dipy.reconst.tests.test_dsi import sticks_and_ball_dummies
from dipy.sims.voxel import sticks_and_ball


def test_gqi():
    # load repulsion 724 sphere
    sphere = default_sphere
    # load icosahedron sphere
    sphere2 = create_unit_sphere(recursion_level=5)
    btable = np.loadtxt(get_fnames(name="dsi515btable"))
    bvals = btable[:, 0]
    bvecs = btable[:, 1:]
    gtab = gradient_table(bvals, bvecs=bvecs)
    data, golden_directions = sticks_and_ball(
        gtab, d=0.0015, S0=100, angles=[(0, 0), (90, 0)], fractions=[50, 50], snr=None
    )
    gq = GeneralizedQSamplingModel(gtab, method="gqi2", sampling_length=1.4)

    # repulsion724
    gqfit = gq.fit(data)
    odf = gqfit.odf(sphere)
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, golden_directions), 2, 1)

    # 5 subdivisions
    gqfit = gq.fit(data)
    odf2 = gqfit.odf(sphere2)
    directions, values, indices = peak_directions(
        odf2, sphere2, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, golden_directions), 2, 1)

    sb_dummies = sticks_and_ball_dummies(gtab)
    for sbd in sb_dummies:
        data, golden_directions = sb_dummies[sbd]
        odf = gq.fit(data).odf(sphere2)
        directions, values, indices = peak_directions(
            odf, sphere2, relative_peak_threshold=0.35, min_separation_angle=25
        )
        if len(directions) <= 3:
            assert_equal(len(directions), len(golden_directions))
        if len(directions) > 3:
            assert_equal(gfa(odf) < 0.1, True)


def test_mvoxel_gqi():
    data, gtab = dsi_voxels()
    sphere = get_sphere(name="symmetric724")

    gq = GeneralizedQSamplingModel(gtab, method="standard")
    gqfit = gq.fit(data)
    all_odfs = gqfit.odf(sphere)

    # Check that the first and last voxels each have 2 peaks
    odf = all_odfs[0, 0, 0]
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(directions.shape[0], 2)
    odf = all_odfs[-1, -1, -1]
    directions, values, indices = peak_directions(
        odf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(directions.shape[0], 2)



import importlib
import sys
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.data import get_fnames
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
have_nn = have_tf or have_torch
BACKENDS = [
    backend
    for backend, available in [("tensorflow", have_tf), ("torch", have_torch)]
    if available
]


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_default_weights(monkeypatch):
    file_path = get_fnames(name="evac_test_data")
    input_arr = np.load(file_path)["input"][0]
    output_arr = np.load(file_path)["output"][0]

    for backend in BACKENDS:
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            evac = dipy_nn.evac

        evac_model = evac.EVACPlus()
        results_arr = evac_model.predict(input_arr, np.eye(4), return_prob=True)
        npt.assert_almost_equal(results_arr, output_arr, decimal=2)


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_default_weights_batch(monkeypatch):
    file_path = get_fnames(name="evac_test_data")
    input_arr = np.load(file_path)["input"]
    output_arr = np.load(file_path)["output"]
    input_arr = list(input_arr)

    for backend in BACKENDS:
        print(backend)
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            evac = dipy_nn.evac

        evac_model = evac.EVACPlus()
        fake_affine = np.array([np.eye(4), np.eye(4)])
        fake_voxsize = np.ones((2, 3))
        results_arr = evac_model.predict(
            input_arr, fake_affine, voxsize=fake_voxsize, batch_size=2, return_prob=True
        )
        npt.assert_almost_equal(results_arr, output_arr, decimal=2)


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_T1_error(monkeypatch):
    for backend in BACKENDS:
        print(backend)
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            evac = dipy_nn.evac

        T1 = np.ones((3, 32, 32, 32))
        evac_model = evac.EVACPlus()
        fake_affine = np.array([np.eye(4), np.eye(4), np.eye(4)])
        npt.assert_raises(ValueError, evac_model.predict, T1, fake_affine)



"""Read test or example data."""

import gzip
import json
from os.path import dirname, exists, join as pjoin
import pickle

import numpy as np
from scipy.sparse import load_npz

from dipy.core.gradients import GradientTable, gradient_table
from dipy.core.sphere import HemiSphere, Sphere
from dipy.data.fetcher import (
    fetch_30_bundle_atlas_hcp842,
    fetch_bundle_atlas_hcp842,
    fetch_bundle_fa_hcp,
    fetch_bundle_warp_dataset,
    fetch_bundles_2_subjects,
    fetch_cenir_multib,
    fetch_cfin_multib,
    fetch_deepn4_test,
    fetch_deepn4_tf_weights,
    fetch_deepn4_torch_weights,
    fetch_disco1_dataset,
    fetch_disco2_dataset,
    fetch_disco3_dataset,
    fetch_disco_dataset,
    fetch_evac_test,
    fetch_evac_tf_weights,
    fetch_evac_torch_weights,
    fetch_gold_standard_io,
    fetch_hbn,
    fetch_hcp,
    fetch_isbi2013_2shell,
    fetch_ivim,
    fetch_mni_template,
    fetch_ptt_minimal_dataset,
    fetch_resdnn_tf_weights,
    fetch_resdnn_torch_weights,
    fetch_scil_b0,
    fetch_sherbrooke_3shell,
    fetch_stanford_hardi,
    fetch_stanford_labels,
    fetch_stanford_pve_maps,
    fetch_stanford_t1,
    fetch_stanford_tracks,
    fetch_syn_data,
    fetch_synb0_test,
    fetch_synb0_weights,
    fetch_taiwan_ntu_dsi,
    fetch_target_tractogram_hcp,
    fetch_tissue_data,
    get_bundle_atlas_hcp842,
    get_fnames,
    get_target_tractogram_hcp,
    get_two_hcp842_bundles,
    read_DiB_70_lte_pte_ste,
    read_DiB_217_lte_pte_ste,
    read_bundles_2_subjects,
    read_cenir_multib,
    read_cfin_dwi,
    read_cfin_t1,
    read_five_af_bundles,
    read_isbi2013_2shell,
    read_ivim,
    read_mni_template,
    read_qte_lte_pte,
    read_scil_b0,
    read_sherbrooke_3shell,
    read_stanford_hardi,
    read_stanford_labels,
    read_stanford_pve_maps,
    read_stanford_t1,
    read_syn_data,
    read_taiwan_ntu_dsi,
    read_tissue_data,
)
from dipy.io.image import load_nifti
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import relist_streamlines
from dipy.utils.arrfuncs import as_native_array

__all__ = [
    "fetch_30_bundle_atlas_hcp842",
    "fetch_bundle_atlas_hcp842",
    "fetch_bundle_fa_hcp",
    "fetch_bundle_warp_dataset",
    "fetch_bundles_2_subjects",
    "fetch_cenir_multib",
    "fetch_cfin_multib",
    "fetch_deepn4_test",
    "fetch_deepn4_tf_weights",
    "fetch_deepn4_torch_weights",
    "fetch_disco1_dataset",
    "fetch_disco2_dataset",
    "fetch_disco3_dataset",
    "fetch_disco_dataset",
    "fetch_evac_test",
    "fetch_evac_tf_weights",
    "fetch_evac_torch_weights",
    "fetch_gold_standard_io",
    "fetch_hbn",
    "fetch_hcp",
    "fetch_isbi2013_2shell",
    "fetch_ivim",
    "fetch_mni_template",
    "fetch_ptt_minimal_dataset",
    "fetch_resdnn_tf_weights",
    "fetch_resdnn_torch_weights",
    "fetch_scil_b0",
    "fetch_sherbrooke_3shell",
    "fetch_stanford_hardi",
    "fetch_stanford_labels",
    "fetch_stanford_pve_maps",
    "fetch_stanford_t1",
    "fetch_stanford_tracks",
    "fetch_syn_data",
    "fetch_synb0_test",
    "fetch_synb0_weights",
    "fetch_taiwan_ntu_dsi",
    "fetch_target_tractogram_hcp",
    "fetch_tissue_data",
    "get_bundle_atlas_hcp842",
    "get_fnames",
    "get_target_tractogram_hcp",
    "get_two_hcp842_bundles",
    "read_DiB_70_lte_pte_ste",
    "read_DiB_217_lte_pte_ste",
    "read_bundles_2_subjects",
    "read_cenir_multib",
    "read_cfin_dwi",
    "read_cfin_t1",
    "read_five_af_bundles",
    "read_isbi2013_2shell",
    "read_ivim",
    "read_mni_template",
    "read_qte_lte_pte",
    "read_scil_b0",
    "read_sherbrooke_3shell",
    "read_stanford_hardi",
    "read_stanford_labels",
    "read_stanford_pve_maps",
    "read_stanford_t1",
    "read_syn_data",
    "read_taiwan_ntu_dsi",
    "read_tissue_data",
]


def loads_compat(byte_data):
    return pickle.loads(byte_data, encoding="latin1")


DATA_DIR = pjoin(dirname(__file__), "files")
SPHERE_FILES = {
    "symmetric362": pjoin(DATA_DIR, "evenly_distributed_sphere_362.npz"),
    "symmetric642": pjoin(DATA_DIR, "evenly_distributed_sphere_642.npz"),
    "symmetric724": pjoin(DATA_DIR, "evenly_distributed_sphere_724.npz"),
    "repulsion724": pjoin(DATA_DIR, "repulsion724.npz"),
    "repulsion100": pjoin(DATA_DIR, "repulsion100.npz"),
    "repulsion200": pjoin(DATA_DIR, "repulsion200.npz"),
}


class DataError(Exception):
    pass


@warning_for_keywords()
def get_sim_voxels(*, name="fib1"):
    """provide some simulated voxel data

    Parameters
    ----------
    name : str, which file?
        'fib0', 'fib1' or 'fib2'

    Returns
    -------
    dix : dictionary, where dix['data'] returns a 2d array
        where every row is a simulated voxel with different orientation

    Examples
    --------
    >>> from dipy.data import get_sim_voxels
    >>> sv=get_sim_voxels(name='fib1')
    >>> sv['data'].shape == (100, 102)
    True
    >>> sv['fibres']
    '1'
    >>> sv['gradients'].shape == (102, 3)
    True
    >>> sv['bvals'].shape == (102,)
    True
    >>> sv['snr']
    '60'
    >>> sv2=get_sim_voxels(name='fib2')
    >>> sv2['fibres']
    '2'
    >>> sv2['snr']
    '80'

    Notes
    -----
    These sim voxels were provided by M.M. Correia using Rician noise.

    """
    if name == "fib0":
        fname = pjoin(DATA_DIR, "fib0.pkl.gz")
    if name == "fib1":
        fname = pjoin(DATA_DIR, "fib1.pkl.gz")
    if name == "fib2":
        fname = pjoin(DATA_DIR, "fib2.pkl.gz")
    return loads_compat(gzip.open(fname, "rb").read())


@warning_for_keywords()
def get_skeleton(*, name="C1"):
    """Provide skeletons generated from Local Skeleton Clustering (LSC).

    Parameters
    ----------
    name : str, 'C1' or 'C3'

    Returns
    -------
    dix : dictionary

    Examples
    --------
    >>> from dipy.data import get_skeleton
    >>> C=get_skeleton(name='C1')
    >>> len(C.keys())
    117
    >>> for c in C: break
    >>> sorted(C[c].keys())
    ['N', 'hidden', 'indices', 'most']

    """
    if name == "C1":
        fname = pjoin(DATA_DIR, "C1.pkl.gz")
    if name == "C3":
        fname = pjoin(DATA_DIR, "C3.pkl.gz")
    return loads_compat(gzip.open(fname, "rb").read())


@warning_for_keywords()
def get_sphere(*, name="symmetric362"):
    """provide triangulated spheres

    Parameters
    ----------
    name : str
        which sphere - one of:
        * 'symmetric362'
        * 'symmetric642'
        * 'symmetric724'
        * 'repulsion724'
        * 'repulsion100'
        * 'repulsion200'

    Returns
    -------
    sphere : a dipy.core.sphere.Sphere class instance

    Examples
    --------
    >>> import numpy as np
    >>> from dipy.data import get_sphere
    >>> sphere = get_sphere(name="symmetric362")
    >>> verts, faces = sphere.vertices, sphere.faces
    >>> verts.shape == (362, 3)
    True
    >>> faces.shape == (720, 3)
    True
    >>> verts, faces = get_sphere(name="not a sphere name") #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    DataError: No sphere called "not a sphere name"

    """  # noqa: E501
    fname = SPHERE_FILES.get(name)
    if fname is None:
        raise DataError(f'No sphere called "{name}"')
    res = np.load(fname)
    # Set to native byte order to avoid errors in compiled routines for
    # big-endian platforms, when using these spheres.
    return Sphere(
        xyz=as_native_array(res["vertices"]), faces=as_native_array(res["faces"])
    )


default_sphere = HemiSphere.from_sphere(get_sphere(name="repulsion724"))
small_sphere = HemiSphere.from_sphere(get_sphere(name="symmetric362"))


def _gradient_from_file(filename):
    """Reads a gradient file saved as a text file compatible with np.loadtxt
    and saved in the dipy data directory"""

    def gtab_getter():
        gradfile = pjoin(DATA_DIR, filename)
        grad = np.loadtxt(gradfile, delimiter=",")
        gtab = GradientTable(grad)
        return gtab

    return gtab_getter


get_3shell_gtab = _gradient_from_file("gtab_3shell.txt")
get_isbi2013_2shell_gtab = _gradient_from_file("gtab_isbi2013_2shell.txt")
get_gtab_taiwan_dsi = _gradient_from_file("gtab_taiwan_dsi.txt")


def dsi_voxels():
    fimg, fbvals, fbvecs = get_fnames(name="small_101D")
    bvals = np.loadtxt(fbvals)
    bvecs = np.loadtxt(fbvecs).T
    data, _ = load_nifti(fimg)
    gtab = gradient_table(bvals, bvecs=bvecs)
    return data, gtab


def dsi_deconv_voxels():
    from dipy.sims.voxel import sticks_and_ball

    gtab = gradient_table(np.loadtxt(get_fnames(name="dsi515btable")))
    data = np.zeros((2, 2, 2, 515))
    for ix in range(2):
        for iy in range(2):
            for iz in range(2):
                data[ix, iy, iz], _ = sticks_and_ball(
                    gtab,
                    d=0.0015,
                    S0=1.0,
                    angles=[(0, 0), (90, 0)],
                    fractions=[50, 50],
                    snr=None,
                )
    return data, gtab


def mrtrix_spherical_functions():
    """Spherical functions represented by spherical harmonic coefficients and
    evaluated on a discrete sphere.

    Returns
    -------
    func_coef : array (2, 3, 4, 45)
        Functions represented by the coefficients associated with the
        mrtrix spherical harmonic basis of maximal order (l) 8.
    func_discrete : array (2, 3, 4, 81)
        Functions evaluated on `sphere`.
    sphere : Sphere
        The discrete sphere, points on the surface of a unit sphere, used to
        evaluate the functions.

    Notes
    -----
    These coefficients were obtained by using the dwi2SH command of mrtrix.

    """
    func_discrete, _ = load_nifti(pjoin(DATA_DIR, "func_discrete.nii.gz"))
    func_coef, _ = load_nifti(pjoin(DATA_DIR, "func_coef.nii.gz"))
    gradients = np.loadtxt(pjoin(DATA_DIR, "sphere_grad.txt"))
    # gradients[0] and the first volume of func_discrete,
    # func_discrete[..., 0], are associated with the b=0 signal.
    # gradients[:, 3] are the b-values for each gradient/volume.
    sphere = Sphere(xyz=gradients[1:, :3])
    return func_coef, func_discrete[..., 1:], sphere


dipy_cmaps = None


def get_cmap(name):
    """Make a callable, similar to maptlotlib.pyplot.get_cmap."""
    global dipy_cmaps
    if dipy_cmaps is None:
        filename = pjoin(DATA_DIR, "dipy_colormaps.json")
        with open(filename) as f:
            dipy_cmaps = json.load(f)

    desc = dipy_cmaps.get(name)
    if desc is None:
        return None

    def simple_cmap(v):
        """Emulates matplotlib colormap callable"""
        rgba = np.ones((len(v), 4))
        for i, color in enumerate(("red", "green", "blue")):
            x, y0, y1 = zip(*desc[color])
            # Matplotlib allows more complex colormaps, but for users who do
            # not have Matplotlib dipy makes a few simple colormaps available.
            # These colormaps are simple because y0 == y1, and therefore we
            # ignore y1 here.
            rgba[:, i] = np.interp(v, x, y0)
        return rgba

    return simple_cmap


def two_cingulum_bundles():
    fname = get_fnames(name="cb_2")
    res = np.load(fname)
    cb1 = relist_streamlines(res["points"], res["offsets"])
    cb2 = relist_streamlines(res["points2"], res["offsets2"])
    return cb1, cb2


def matlab_life_results():
    matlab_rmse = np.load(pjoin(DATA_DIR, "life_matlab_rmse.npy"))
    matlab_weights = np.load(pjoin(DATA_DIR, "life_matlab_weights.npy"))
    return matlab_rmse, matlab_weights


@warning_for_keywords()
def load_sdp_constraints(model_name, *, order=None):
    """Import semidefinite programming constraint matrices for different models.

    Generated as described for example in :footcite:p:`DelaHaije2020`.

    Parameters
    ----------
    model_name : string
        A string identifying the model that is to be constrained.
    order : unsigned int, optional
        A non-negative integer that represent the order or instance of the
        model.
        Default: None.

    Returns
    -------
    sdp_constraints : array
        Constraint matrices

    Notes
    -----
    The constraints will be loaded from a file named 'id_constraint_order.npz'.

    References
    ----------
    .. footbibliography::

    """

    file = model_name + "_constraint"
    if order is not None:
        file += f"_{str(order)}"
    if order is None:
        file += "_SC"
    file += ".npz"

    path = pjoin(DATA_DIR, file)

    if not exists(path):
        raise ValueError(f"Constraints file '{file}' not found.")

    try:
        array = load_npz(path)
        n, x = array.shape
        sdp_constraints = [array[i * x : (i + 1) * x] for i in range(n // x)]
        return sdp_constraints
    except Exception as e:
        raise ValueError(f"Failed to read constraints file '{file}'.") from e



"""
==================================================================
Patch2Self: Self-Supervised Denoising via Statistical Independence
==================================================================

Patch2Self :footcite:p:`Fadnavis2020` and :footcite:p:`Fadnavis2024` is  a self-supervised learning method for
denoising DWI data, which uses the entire volume to learn a full-rank locally
linear denoiser for that volume. By taking advantage of the oversampled q-space
of DWI data, Patch2Self can separate structure from noise without requiring an
explicit model for either.

Classical denoising algorithms such as Local PCA :footcite:p:`Manjon2013`,
:footcite:p:`Veraart2016b`, Non-local Means :footcite:p:`Coupe2008`, Total
Variation Norm :footcite:p:`Knoll2011`, etc. assume certain properties on the
signal structure. Patch2Self *does not* make any such assumption on the signal,
instead using the fact that the noise across different 3D volumes of the DWI
signal originates from random fluctuations in the acquired signal.

Since Patch2Self only relies on the randomness of the noise, it can be applied
at any step in the pre-processing pipeline. The design of Patch2Self is such
that it can work on any type of diffusion data/ any body part without
requiring a noise estimation or assumptions on the type of noise (such as its
distribution).

The Patch2Self Framework:

.. _patch2self:
.. image:: https://github.com/dipy/dipy_data/blob/master/Patch2Self_Framework.PNG?raw=true
   :width: 70 %
   :align: center

The above figure demonstrates the working of Patch2Self. The idea is to build
a new regressor for denoising each 3D volume of the 4D diffusion data. This is
done in the following 2 phases:

(A) Self-supervised training: First, we extract 3D Patches from all the ``n``
volumes and hold out the target volume to denoise. Each patch from the rest of
the ``(n-1)`` volumes predicts the center voxel of the corresponding patch in
the target volume.

This is done by using the self-supervised loss:

.. math::

    \\mathcal{L}\\left(\\Phi_{J}\\right)=\\mathbb{E}\\left\\|\\Phi_{J}\\
     \\left(Y_{*, *,-j}\\right)-Y_{*, 0, j}\\right\\|^{2}

(B) Prediction: The same 'n-1' volumes which were used in the training are now
fed into the regressor :math:`\\Phi` built in phase (A). The prediction is a
denoised version of held-out volume.

Note: The volume to be denoised is merely used as the target in the training
phase. But is not used in the training set for (A) nor is used to predict the
denoised output in (B).

Let's load the necessary modules:
"""  # noqa: E501

import matplotlib.pyplot as plt
import numpy as np

from dipy.data import get_fnames
from dipy.denoise.patch2self import patch2self
from dipy.io.image import load_nifti, save_nifti

###############################################################################
# Now let's load an example dataset and denoise it with Patch2Self. Patch2Self
# does not require noise estimation and should work with any kind of diffusion
# data.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
data, affine = load_nifti(hardi_fname)
bvals = np.loadtxt(hardi_bval_fname)
denoised_arr = patch2self(
    data,
    bvals,
    model="ols",
    shift_intensity=True,
    clip_negative_vals=False,
    b0_threshold=50,
)

###############################################################################
# The above parameters should give optimal denoising performance for
# Patch2Self. The ordinary least squares regression ``(model='ols')`` tends to
# be a little slower depending on the size of the data. In that case, please
# consider switching to ridge regression ``(model='ridge')``.
#
# Please do note that sometimes using ridge regression can hamper the
# performance of Patch2Self. If so, please use ``model='ols'``.
#
# The array ``denoised_arr`` contains the denoised output obtained from
# Patch2Self.
#
# .. note::
#
#    Depending on the acquisition, b0 may exhibit signal attenuation or
#    other artefacts that are not ideal for any denoising algorithm. We
#    therefore provide an option to skip denoising b0 volumes in the data.
#    This can be done by using the option `b0_denoising=False` within
#    Patch2Self.
#
#    Please set ``shift_intensity=True`` and ``clip_negative_vals=False`` by
#    default to avoid negative values in the denoised output.
#
#    The ``b0_threshold`` is used to separate the b0 volumes from the DWI
#    volumes. Changing the value of the b0 threshold is needed if the b0 volumes
#    in the ``bval`` file have a value greater than the default
#    ``b0_threshold``.
#
#    The default value of ``b0_threshold`` in DIPY is set to 50. If using data
#    such as HCP 7T, the b0 volumes tend to have a higher b-value (>=50)
#    associated with them in the `bval` file. Please check the b-values for b0s
#    and adjust the ``b0_threshold``` accordingly.
#
# Now let's visualize the output and the residuals obtained from the denoising.

# Gets the center slice and the middle volume of the 4D diffusion data.
sli = data.shape[2] // 2
gra = 60  # pick out a random volume for a particular gradient direction

orig = data[:, :, sli, gra]
den = denoised_arr[:, :, sli, gra]

# computes the residuals
rms_diff = np.sqrt((orig - den) ** 2)

fig1, ax = plt.subplots(1, 3, figsize=(12, 6), subplot_kw={"xticks": [], "yticks": []})

fig1.subplots_adjust(hspace=0.3, wspace=0.05)

ax.flat[0].imshow(orig.T, cmap="gray", interpolation="none", origin="lower")
ax.flat[0].set_title("Original")
ax.flat[1].imshow(den.T, cmap="gray", interpolation="none", origin="lower")
ax.flat[1].set_title("Denoised Output")
ax.flat[2].imshow(rms_diff.T, cmap="gray", interpolation="none", origin="lower")
ax.flat[2].set_title("Residuals")

fig1.savefig("denoised_patch2self.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Patch2Self preserved anatomical detail. This can be visually verified by
# inspecting the residuals obtained above. Since we do not see any structure in
# the difference residuals, it is clear that it preserved the underlying signal
# structure and got rid of the stochastic noise.
#
#
# Below we show how the denoised data can be saved.

save_nifti("denoised_patch2self.nii.gz", denoised_arr, affine)

###############################################################################
# You can also use Patch2Self version 1 to denoise the data by using version
# argument. The default version is set to 3. To use version 1, you can call
# Patch2Self as follows::
#
#    patch2self(data, bvals, version=1)
#
# Lastly, one can also use Patch2Self in batches if the number of gradient
# directions is very high (>=200 volumes). For instance, if the data has 300
# volumes, one can split the data into 2 batches, (150 directions each) and
# still get the same denoising performance. One can run Patch2Self
# using::
#
#    denoised_batch1 = patch2self(data[..., :150], bvals[:150])
#    denoised_batch2 = patch2self(data[..., 150:], bvals[150:])
#
# After doing this, the 2 denoised batches can be merged as follows::
#
#    denoised_p2s = np.concatenate((denoised_batch1, denoised_batch2), axis=3)
#
# One can also consider using the above batching approach to denoise each
# shell separately if working with multi-shell data.
#
#
# References
# ----------
#
# .. footbibliography::



import warnings

import numpy as np
from scipy.ndimage import label


def remove_holes_and_islands(binary_img, *, slice_wise=False):
    """
    Remove any small mask chunks or holes
    that could be in the segmentation output.

    Parameters
    ----------
    binary_img : np.ndarray
        Binary image
    slice_wise : bool, optional
        Whether to run slice wise background correction as well

    Returns
    -------
    largest_img : np.ndarray
    """
    if len(np.unique(binary_img)) != 2:
        warnings.warn(
            "The mask is not binary. \
                      Returning the original mask",
            UserWarning,
            stacklevel=2,
        )
        return binary_img

    largest_img = np.zeros_like(binary_img)
    chunks, _ = label(np.abs(1 - binary_img))
    u, c = np.unique(chunks[chunks != 0], return_counts=True)
    if len(u) == 0:
        warnings.warn(
            "The mask has no background. \
                      Returning the original mask",
            UserWarning,
            stacklevel=2,
        )
        return binary_img
    target = u[np.argmax(c)]

    largest_img = np.where(chunks == target, 0, 1)

    chunks, _ = label(largest_img)
    u, c = np.unique(chunks[chunks != 0], return_counts=True)
    if len(u) == 0:
        warnings.warn(
            "The mask has no foreground. \
                      Returning the original mask",
            UserWarning,
            stacklevel=2,
        )
        return binary_img
    target = u[np.argmax(c)]

    largest_img = np.where(chunks == target, 1, 0)

    if slice_wise:
        for x in range(largest_img.shape[0]):
            chunks, n_chunk = label(np.abs(1 - largest_img[x]))
            u, c = np.unique(chunks[chunks != 0], return_counts=True)
            target = u[np.argmax(c)]
            largest_img[x] = np.where(chunks == target, 0, 1)
        for y in range(largest_img.shape[1]):
            chunks, n_chunk = label(np.abs(1 - largest_img[:, y]))
            u, c = np.unique(chunks[chunks != 0], return_counts=True)
            target = u[np.argmax(c)]
            largest_img[:, y] = np.where(chunks == target, 0, 1)
        for z in range(largest_img.shape[2]):
            chunks, n_chunk = label(np.abs(1 - largest_img[..., z]))
            u, c = np.unique(chunks[chunks != 0], return_counts=True)
            target = u[np.argmax(c)]
            largest_img[..., z] = np.where(chunks == target, 0, 1)

    return largest_img



import numpy as np

from dipy.segment.utils import remove_holes_and_islands
from dipy.testing.decorators import set_random_number_generator


@set_random_number_generator()
def test_remove_holes_and_islands(rng=None):
    # inefficient but more readable
    temp = rng.choice([0, 1], size=(40, 40, 40), p=[0.8, 0.2])
    temp[6:34, 6:34, 6:34] = 0
    temp[7:33, 7:33, 7:33] = 1
    temp[8:32, 8:32, 8:32] = rng.choice([0, 1], size=(24, 24, 24), p=[0.2, 0.8])
    output = remove_holes_and_islands(temp)
    ground_truth = np.zeros((40, 40, 40))
    ground_truth[7:33, 7:33, 7:33] = 1
    np.testing.assert_equal(output, ground_truth)


def test_remove_holes_and_islands_warnings():
    # Not binary test
    non_binary_img = np.concatenate(
        [np.zeros((30, 30, 10)), np.ones((30, 30, 10)), np.ones((30, 30, 10)) * 2],
        axis=-1,
    )
    np.testing.assert_warns(UserWarning, remove_holes_and_islands, non_binary_img)

    # No background test
    no_background_img = np.ones((40, 40, 40))
    np.testing.assert_warns(UserWarning, remove_holes_and_islands, no_background_img)

    # No foreground test
    no_foreground_img = np.zeros((40, 40, 40))
    np.testing.assert_warns(UserWarning, remove_holes_and_islands, no_foreground_img)



import numpy as np

from dipy.core.geometry import normalized_vector
from dipy.core.gradients import gradient_table
from dipy.stats.qc import neighboring_dwi_correlation

rng = np.random.default_rng()


def create_test_data(test_r, cube_size, mask_size, num_dwi_vols, num_b0s):
    """Create testing data with a known neighbor structure and a known NDC.

    The b>0 images have 2 images per shell, separated by a very small angle,
    guaranteeing they will be neighbors. The within-mask data is filled with
    random data with correlation value of approximately ``test_r``.

    Parameters
    ----------

    test_r: float
        The approximate NDC that the simulated data should have
    cube_size: int
        The simulated data will be a cube with this many voxels per dim
    mask_size: int
        A cubic "brain" is this size per side and filled with data. Must
        be less than ``cube_size``
    num_dwi_vols: int
        The number of b>0 images to simulate. Must be even to ensure we
        can make known neighbors
    num_b0s: int
        The number of b=0 images to prepend to the b>0 images

    Returns
    -------

    real_r: float
        The ground-truth neighbor correlation of the simulated data
    dwi_data: np.ndarray
        A 4D array containing simulated data
    mask_data: np.ndarray
        A 3D array indicating which voxels in ``dwi_data`` contain
        brain data
    gtab: dipy.core.gradients.GradientTable
        Gradient table with known neighbors

    """

    if not num_dwi_vols % 2 == 0:
        raise Exception("Needs an even number of dwi vols to ensure known neighbors")

    # Create a volume mask
    test_mask = np.zeros((cube_size, cube_size, cube_size))
    test_mask[:mask_size, :mask_size, :mask_size] = 1
    n_voxels_in_mask = mask_size**3

    # 4D Data array
    dwi_data = np.zeros((cube_size, cube_size, cube_size, num_b0s + num_dwi_vols))

    # Create a sampling scheme where we know what volumes will be neighbors
    n_known = num_dwi_vols // 2
    dwi_bvals = (
        np.column_stack([np.arange(n_known) + 1] * 2).flatten(order="C").tolist()
    )
    bvals = np.array([0] * num_b0s + dwi_bvals) * 1000

    # The bvecs will be a straight line with a minor perturbance every other
    ref_vec = np.array([1.0, 0.0, 0.0])
    nbr_vec = normalized_vector(ref_vec + 0.00001)
    bvecs = np.vstack([ref_vec] * num_b0s + [np.vstack([ref_vec, nbr_vec])] * n_known)

    cor = np.ones((2, 2)) * test_r
    np.fill_diagonal(cor, 1)
    L = np.linalg.cholesky(cor)

    known_correlations = []
    for starting_vol in np.arange(n_known) * 2 + num_b0s:
        uncorrelated = rng.standard_normal((2, n_voxels_in_mask))
        correlated = np.dot(L, uncorrelated)

        dwi_data[:, :, :, starting_vol][test_mask > 0] = correlated[0]
        dwi_data[:, :, :, starting_vol + 1][test_mask > 0] = correlated[1]

        known_correlations += [np.corrcoef(correlated)[0, 1]] * 2

    gtab = gradient_table(bvals, bvecs=bvecs, b0_threshold=50)

    return np.mean(known_correlations), dwi_data, test_mask, gtab


def test_neighboring_dwi_correlation():
    """Test NDC under various conditions."""

    # Test data with b=0s, low correlation, using mask
    real_r, dwi_data, mask, gtab = create_test_data(
        test_r=0.3, cube_size=10, mask_size=6, num_dwi_vols=10, num_b0s=2
    )
    estimated_ndc = neighboring_dwi_correlation(dwi_data, gtab, mask=mask)
    assert np.allclose(real_r, estimated_ndc)

    maskless_ndc = neighboring_dwi_correlation(dwi_data, gtab)
    assert maskless_ndc != real_r

    # Try with no b=0s
    real_r, dwi_data, mask, gtab = create_test_data(
        test_r=0.3, cube_size=10, mask_size=6, num_dwi_vols=10, num_b0s=0
    )
    estimated_ndc = neighboring_dwi_correlation(dwi_data, gtab, mask=mask)
    assert np.allclose(real_r, estimated_ndc)

    # Try with realistic correlation value
    real_r, dwi_data, mask, gtab = create_test_data(
        test_r=0.8, cube_size=10, mask_size=6, num_dwi_vols=10, num_b0s=2
    )
    estimated_ndc = neighboring_dwi_correlation(dwi_data, gtab, mask=mask)
    assert np.allclose(real_r, estimated_ndc)

    # Try with a bigger volume, lower correlation
    real_r, dwi_data, mask, gtab = create_test_data(
        test_r=0.5, cube_size=100, mask_size=49, num_dwi_vols=160, num_b0s=2
    )
    estimated_ndc = neighboring_dwi_correlation(dwi_data, gtab, mask=mask)
    assert np.allclose(real_r, estimated_ndc)



"""
================================================
Reconstruction with Multi-Shell Multi-Tissue CSD
================================================

This example shows how to use Multi-Shell Multi-Tissue Constrained Spherical
Deconvolution (MSMT-CSD) introduced by :footcite:t:`Jeurissen2014`. This
tutorial goes through the steps involved in implementing the method.

This method provides improved White Matter(WM), Grey Matter (GM), and
Cerebrospinal fluid (CSF) volume fraction maps, which is otherwise
overestimated in the standard CSD (SSST-CSD). This is done by using b-value
dependencies of the different tissue types to estimate ODFs. This method thus
extends the SSST-CSD introduced in :footcite:p:`Tournier2007`.

The reconstruction of the fiber orientation distribution function
(fODF) in MSMT-CSD involves the following steps:

    1. Generate a mask using Median Otsu (optional step)
    2. Denoise the data using MP-PCA (optional step)
    3. Generate  Anisotropic Powermap (if T1 unavailable)
    4. Fit DTI model to the data
    5. Tissue Classification (needs to be at least two classes of tissues)
    6. Estimation of the fiber response function
    7. Use the response function to reconstruct the fODF

First, we import all the modules we need from dipy as follows:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table, unique_bvals_tolerance
from dipy.data import get_fnames, get_sphere
from dipy.denoise.localpca import mppca
import dipy.direction.peaks as dp
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
from dipy.reconst.mcsd import (
    MultiShellDeconvModel,
    auto_response_msmt,
    mask_for_response_msmt,
    multi_shell_fiber_response,
    response_from_mask_msmt,
)
import dipy.reconst.shm as shm
from dipy.segment.mask import median_otsu
from dipy.segment.tissue import TissueClassifierHMRF
from dipy.viz import actor, window

sphere = get_sphere(name="symmetric724")

###############################################################################
# For this example, we use fetch to download a multi-shell dataset which was
# kindly provided by Hansen and Jespersen (more details about the data are
# provided in their paper :footcite:p:`Hansen2016a`). The total size of the
# downloaded data is 192 MBytes, however you only need to fetch it once.

fraw, fbval, fbvec, t1_fname = get_fnames(name="cfin_multib")

data, affine = load_nifti(fraw)
bvals, bvecs = read_bvals_bvecs(fbval, fbvec)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# For the sake of simplicity, we only select two non-zero b-values for this
# example.

bvals = gtab.bvals
bvecs = gtab.bvecs

sel_b = np.logical_or(np.logical_or(bvals == 0, bvals == 1000), bvals == 2000)
data = data[..., sel_b]

###############################################################################
# The gradient table is also selected to have the selected b-values (0, 1000
# and 2000)

gtab = gradient_table(bvals[sel_b], bvecs=bvecs[sel_b])

###############################################################################
# We make use of the ``median_otsu`` method to generate the mask for the data
# as follows:

b0_mask, mask = median_otsu(data, median_radius=2, numpass=1, vol_idx=[0, 1])

print(data.shape)

###############################################################################
# As one can see from its shape, the selected data contains a total of 67
# volumes of images corresponding to all the diffusion gradient directions
# of the selected b-values and call the ``mppca`` as follows:

denoised_arr = mppca(data, mask=mask, patch_radius=2)

###############################################################################
# Now we will use the denoised array (``denoised_arr``) obtained from ``mppca``
# in the rest of the steps in the tutorial.
#
# As for the next step, we generate the anisotropic powermap introduced by
# :footcite:t:`DellAcqua2014`. To do so, we make use of the Q-ball Model as
# follows:

qball_model = shm.QballModel(gtab, 8)

###############################################################################
# We generate the peaks from the ``qball_model`` as follows:

peaks = dp.peaks_from_model(
    model=qball_model,
    data=denoised_arr,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    sphere=sphere,
    mask=mask,
)

ap = shm.anisotropic_power(peaks.shm_coeff)

plt.matshow(np.rot90(ap[:, :, 10]), cmap=plt.cm.bone)
plt.savefig("anisotropic_power_map.png")
plt.close()

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Anisotropic Power Map (Axial Slice)

print(ap.shape)

###############################################################################
# The above figure is a visualization of the axial slice of the Anisotropic
# Power Map. It can be treated as a pseudo-T1 for classification purposes
# using the Hidden Markov Random Fields (HMRF) classifier, if the T1 image
# is not available.
#
# As we can see from the shape of the Anisotropic Power Map, it is 3D and can
# be used for tissue classification using HMRF. The
# HMRF needs the specification of the number of classes. For the case of
# MSMT-CSD the ``nclass`` parameter needs to be ``>=2``. In our case, we set
# it to 3: namely corticospinal fluid (csf), white matter (wm) and gray
# matter (gm).

nclass = 3

###############################################################################
# Then, the smoothness factor of the segmentation. Good performance is achieved
# with values between 0 and 0.5.

beta = 0.1

###############################################################################
# We then call the ``TissueClassifierHMRF`` with the parameters specified as
# above:

hmrf = TissueClassifierHMRF()
initial_segmentation, final_segmentation, PVE = hmrf.classify(ap, nclass, beta)

###############################################################################
# Then, we get the tissues segmentation from the final_segmentation.

csf = np.where(final_segmentation == 1, 1, 0)
gm = np.where(final_segmentation == 2, 1, 0)
wm = np.where(final_segmentation == 3, 1, 0)

###############################################################################
# Now, we want the response function for each of the three tissues and for each
# bvalues. This can be achieved in two different ways. If the case that tissue
# segmentation is available or that one wants to see the tissue masks used to
# compute the response functions, a combination of the functions
# ``mask_for_response_msmt`` and ``response_from_mask`` is needed.
#
# The ``mask_for_response_msmt`` function will return a mask of voxels within a
# cuboid ROI and that meet some threshold constraints, for each tissue and
# bvalue. More precisely, the WM mask must have a FA value above a given
# threshold. The GM mask and CSF mask must have a FA below given thresholds
# and a MD below other thresholds.
#
# Note that for ``mask_for_response_msmt``, the gtab and data should be for
# bvalues under 1200, for optimal tensor fit.

mask_wm, mask_gm, mask_csf = mask_for_response_msmt(
    gtab,
    data,
    roi_radii=10,
    wm_fa_thr=0.7,
    gm_fa_thr=0.3,
    csf_fa_thr=0.15,
    gm_md_thr=0.001,
    csf_md_thr=0.0032,
)

###############################################################################
# If one wants to use the previously computed tissue segmentation in addition
# to the threshold method, it is possible by simply multiplying both masks
# together.

mask_wm *= wm
mask_gm *= gm
mask_csf *= csf

###############################################################################
# The masks can also be used to calculate the number of voxels for each tissue.

nvoxels_wm = np.sum(mask_wm)
nvoxels_gm = np.sum(mask_gm)
nvoxels_csf = np.sum(mask_csf)

print(nvoxels_wm)

###############################################################################
# Then, the ``response_from_mask`` function will return the msmt response
# functions using precalculated tissue masks.

response_wm, response_gm, response_csf = response_from_mask_msmt(
    gtab, data, mask_wm, mask_gm, mask_csf
)

###############################################################################
# Note that we can also get directly the response functions by calling the
# ``auto_response_msmt`` function, which internally calls
# ``mask_for_response_msmt`` followed by ``response_from_mask``. By doing so,
# we don't have access to the masks and we might have problems with high
# bvalues tensor fit.

auto_response_wm, auto_response_gm, auto_response_csf = auto_response_msmt(
    gtab, data, roi_radii=10
)

###############################################################################
# As we can see below, adding the tissue segmentation can change the results
# of the response functions.

print("Responses")
print(response_wm)
print(response_gm)
print(response_csf)
print("Auto responses")
print(auto_response_wm)
print(auto_response_gm)
print(auto_response_csf)

###############################################################################
# At this point, there are two options on how to use those response functions.
# We want to create a MultiShellDeconvModel, which takes a response function as
# input. This response function can either be directly in the current format,
# or it can be a MultiShellResponse format, as produced by the
# ``multi_shell_fiber_response`` method. This function assumes a 3 compartments
# model (wm, gm, csf) and takes one response function per tissue per bvalue. It
# is important to note that the bvalues must be unique for this function.

ubvals = unique_bvals_tolerance(gtab.bvals)
response_mcsd = multi_shell_fiber_response(
    sh_order_max=8,
    bvals=ubvals,
    wm_rf=response_wm,
    gm_rf=response_gm,
    csf_rf=response_csf,
)

###############################################################################
# As mentioned, we can also build the model directly and it will call
# ``multi_shell_fiber_response`` internally. Important note here, the function
# ``unique_bvals_tolerance`` is used to keep only unique bvalues from the gtab
# given to the model, as input for ``multi_shell_fiber_response``. This may
# introduce differences between the calculated response of each method,
# depending on the bvalues given to ``multi_shell_fiber_response`` externally.

response = np.array([response_wm, response_gm, response_csf])
mcsd_model_simple_response = MultiShellDeconvModel(gtab, response, sh_order_max=8)

###############################################################################
# Note that this technique only works for a 3 compartments model (wm, gm, csf).
# If one wants more compartments, a custom MultiShellResponse object must be
# used. It can be inspired by the ``multi_shell_fiber_response`` method.
#
# Now we build the MSMT-CSD model with the ``response_mcsd`` as input. We then
# call the ``fit`` function to fit one slice of the 3D data and visualize it.

mcsd_model = MultiShellDeconvModel(gtab, response_mcsd)
mcsd_fit = mcsd_model.fit(denoised_arr[:, :, 10:11])

###############################################################################
# The volume fractions of tissues for each voxel are also accessible, as well
# as the sh coefficients for all tissues. One can also get each sh tissue
# separately using ``all_shm_coeff`` for each compartment (isotropic) and
# ``shm_coeff`` for white matter.

vf = mcsd_fit.volume_fractions
sh_coeff = mcsd_fit.all_shm_coeff
csf_sh_coeff = sh_coeff[..., 0]
gm_sh_coeff = sh_coeff[..., 1]
wm_sh_coeff = mcsd_fit.shm_coeff

###############################################################################
# The model allows one to predict a signal from sh coefficients. There are two
# ways of doing this.

mcsd_pred = mcsd_fit.predict()
mcsd_pred = mcsd_model.predict(mcsd_fit.all_shm_coeff)

###############################################################################
# From the fit obtained in the previous step, we generate the ODFs which can be
# visualized as follows:

mcsd_odf = mcsd_fit.odf(sphere)

print("ODF")
print(mcsd_odf.shape)
print(mcsd_odf[40, 40, 0])

fodf_spheres = actor.odf_slicer(
    mcsd_odf, sphere=sphere, scale=1, norm=False, colormap="plasma"
)

interactive = False
scene = window.Scene()
scene.add(fodf_spheres)
scene.reset_camera_tight()

print("Saving illustration as msdodf.png")
window.record(scene=scene, out_path="msdodf.png", size=(600, 600))

if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# MSMT-CSD Peaks and ODFs.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""Random number generation utilities."""

from math import floor
from platform import architecture

import numpy as np

from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def WichmannHill2006(*, ix=100001, iy=200002, iz=300003, it=400004):
    """Wichmann Hill random number generator.

    See :footcite:p:`Wichmann2006` for advice on generating many sequences for
    use together, and on alternative algorithms and codes

    Parameters
    ----------
    ix: int
        First seed value. Should not be null. (default 100001)
    iy: int
        Second seed value. Should not be null. (default 200002)
    iz: int
        Third seed value. Should not be null. (default 300003)
    it: int
        Fourth seed value. Should not be null. (default 400004)

    Returns
    -------
    r_number : float
        pseudo-random number uniformly distributed between [0-1]

    References
    ----------
    .. footbibliography::

    Examples
    --------
    >>> from dipy.core import rng
    >>> N = 1000
    >>> a = [rng.WichmannHill2006() for i in range(N)]

    """
    if not ix or not iy or not iz or not it:
        raise ValueError("A seed value can not be null.")

    if architecture()[0] == "64":
        # If 64 bits are available then the following lines of code will be
        # faster.
        ix = (11600 * ix) % 2147483579
        iy = (47003 * iy) % 2147483543
        iz = (23000 * iz) % 2147483423
        it = (33000 * it) % 2147483123

    else:
        # If only 32 bits are available

        ix = 11600 * (ix % 185127) - 10379 * (ix / 185127)
        iy = 47003 * (ix % 45688) - 10479 * (iy / 45688)
        iz = 23000 * (iz % 93368) - 19423 * (iz / 93368)
        it = 33000 * (it % 65075) - 8123 * (it / 65075)

        if ix < 0:
            ix = ix + 2147483579
        if iy < 0:
            iy = iy + 2147483543
        if iz < 0:
            iz = iz + 2147483423
        if it < 0:
            it = it + 2147483123

    W = ix / 2147483579.0 + iy / 2147483543.0 + iz / 2147483423.0 + it / 2147483123.0

    return W - floor(W)


@warning_for_keywords()
def WichmannHill1982(*, ix=100001, iy=200002, iz=300003):
    """Algorithm AS 183 Appl. Statist. (1982) vol.31, no.2.

    Returns a pseudo-random number rectangularly distributed
    between 0 and 1.   The cycle length is 6.95E+12 (See page 123
    of Applied Statistics (1984) vol.33), not as claimed in the
    original article.

    ix, iy and iz should be set to integer values between 1 and
    30000 before the first entry.

    Integer arithmetic up to 5212632 is required.

    Parameters
    ----------
    ix: int
        First seed value. Should not be null. (default 100001)
    iy: int
        Second seed value. Should not be null. (default 200002)
    iz: int
        Third seed value. Should not be null. (default 300003)

    Returns
    -------
    r_number : float
        pseudo-random number uniformly distributed between [0-1]

    Examples
    --------
    >>> from dipy.core import rng
    >>> N = 1000
    >>> a = [rng.WichmannHill1982() for i in range(N)]

    """
    if not ix or not iy or not iz:
        raise ValueError("A seed value can not be null.")
    ix = (171 * ix) % 30269
    iy = (172 * iy) % 30307
    iz = (170 * iz) % 30323

    """
    If integer arithmetic only up to 30323 (!) is available, the preceding
    3 statements may be replaced by:

    ix = 171 * (ix % 177) -  2 * (ix / 177)
    iy = 172 * (iy % 176) - 35 * (iy / 176)
    iz = 170 * (iz % 178) - 63 * (iz / 178)

    if ix < 0:
        ix = ix + 30269
    if iy < 0:
        iy = iy + 30307
    if iz < 0:
        iz = iz + 30323
    """
    return np.remainder(
        float(ix) / 30269.0 + float(iy) / 30307.0 + float(iz) / 30323.0, 1.0
    )


@warning_for_keywords()
def LEcuyer(*, s1=100001, s2=200002):
    """Return a LEcuyer random number generator.

    Generate uniformly distributed random numbers using the 32-bit
    generator from figure 3 of:

        L'Ecuyer, P. Efficient and portable combined random number
        generators, C.A.C.M., vol. 31, 742-749 & 774-?, June 1988.

    The cycle length is claimed to be 2.30584E+18

    Parameters
    ----------
    s1: int
        First seed value. Should not be null. (default 100001)
    s2: int
        Second seed value. Should not be null. (default 200002)

    Returns
    -------
    r_number : float
        pseudo-random number uniformly distributed between [0-1]

    Examples
    --------
    >>> from dipy.core import rng
    >>> N = 1000
    >>> a = [rng.LEcuyer() for i in range(N)]

    """
    if not s1 or not s2:
        raise ValueError("A seed value can not be null.")
    k = s1 / 53668
    s1 = 40014 * (s1 - k * 53668) - k * 12211
    if s1 < 0:
        s1 = s1 + 2147483563
    k = s2 / 52774
    s2 = 40692 * (s2 - k * 52774) - k * 3791
    if s2 < 0:
        s2 = s2 + 2147483399
    z = s1 - s2
    if z < 0:
        z = z + 2147483562

    return z / 2147483563.0



"""
==============================================================
Denoise images using Adaptive Soft Coefficient Matching (ASCM)
==============================================================

The adaptive soft coefficient matching (ASCM) as described in
:footcite:p:`Coupe2012` is an improved extension of non-local means (NLMEANS)
denoising. ASCM gives a better denoised images from two standard non-local means
denoised versions of the original data with different degrees sharpness. Here,
one denoised input is more "smooth" than the other (the easiest way to achieve
this denoising is use :ref:`non_local_means<sphx_glr_examples_built_preprocessing_denoise_nlmeans.py>`
with two different patch radii).

ASCM involves these basic steps

* Computes wavelet decomposition of the noisy as well as denoised inputs

* Combines the wavelets for the output image in a way that it takes it's
  smoothness (low frequency components) from the input with larger smoothing,
  and the sharp features (high frequency components) from the input with
  less smoothing.

This way ASCM gives us a well denoised output while preserving the sharpness
of the image features.

Let us load the necessary modules
"""  # noqa: E501

from time import time

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.denoise.adaptive_soft_matching import adaptive_soft_matching
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.denoise.non_local_means import non_local_means
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti

###############################################################################
# Choose one of the data from the datasets in dipy_

dwi_fname, dwi_bval_fname, dwi_bvec_fname = get_fnames(name="sherbrooke_3shell")
data, affine = load_nifti(dwi_fname)
bvals, bvecs = read_bvals_bvecs(dwi_bval_fname, dwi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

mask = data[..., 0] > 80
data = data[..., 1]

print("vol size", data.shape)

t = time()

###############################################################################
# In order to generate the two pre-denoised versions of the data we will use
# the :ref:`non_local_means denoining<sphx_glr_examples_built_preprocessing_denoise_nlmeans.py>`   # noqa: E501
# For ``non_local_means`` first we need to estimate the standard deviation of
# the noise. We use N=4 since the Sherbrooke dataset was acquired on a
# 1.5T Siemens scanner with a 4 array head coil.

sigma = estimate_sigma(data, N=4)

###############################################################################
# For the denoised version of the original data which preserves sharper
# features, we perform non-local means with smaller patch size.

den_small = non_local_means(
    data, sigma=sigma, mask=mask, patch_radius=1, block_radius=1, rician=True
)

###############################################################################
# For the denoised version of the original data that implies more smoothing, we
# perform non-local means with larger patch size.

den_large = non_local_means(
    data, sigma=sigma, mask=mask, patch_radius=2, block_radius=1, rician=True
)

###############################################################################
# Now we perform the adaptive soft coefficient matching. Empirically we set the
# adaptive parameter in ascm to be the average of the local noise variance,
# in this case the sigma itself.

den_final = adaptive_soft_matching(data, den_small, den_large, sigma[0])

print("total time", time() - t)

###############################################################################
# To access the quality of this denoising procedure, we plot an axial slice
# of the original data, it's denoised output and residuals.

axial_middle = data.shape[2] // 2

original = data[:, :, axial_middle].T
final_output = den_final[:, :, axial_middle].T
difference = np.abs(final_output.astype(np.float64) - original.astype(np.float64))
difference[~mask[:, :, axial_middle].T] = 0

fig, ax = plt.subplots(1, 3)
ax[0].imshow(original, cmap="gray", origin="lower")
ax[0].set_title("Original")
ax[1].imshow(final_output, cmap="gray", origin="lower")
ax[1].set_title("ASCM output")
ax[2].imshow(difference, cmap="gray", origin="lower")
ax[2].set_title("Residual")
for i in range(3):
    ax[i].set_axis_off()

plt.savefig("denoised_ascm.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the axial slice without (left) and with (middle) ASCM denoising.
#
#
# From the above figure we can see that the residual is really uniform in
# nature which dictates that ASCM denoises the data while preserving the
# sharpness of the features. Now, we are Saving the entire denoised output in
# ``denoised_ascm.nii.gz`` file.

save_nifti("denoised_ascm.nii.gz", den_final, affine)

###############################################################################
# For comparison propose we also plot the outputs of the ``non_local_means``
# (both with the larger as well as with the smaller patch radius) with the ASCM
# output.

fig, ax = plt.subplots(1, 4)
ax[0].imshow(original, cmap="gray", origin="lower")
ax[0].set_title("Original")
ax[1].imshow(
    den_small[..., axial_middle].T, cmap="gray", origin="lower", interpolation="none"
)
ax[1].set_title("NLMEANS small")
ax[2].imshow(
    den_large[..., axial_middle].T, cmap="gray", origin="lower", interpolation="none"
)
ax[2].set_title("NLMEANS large")
ax[3].imshow(final_output, cmap="gray", origin="lower", interpolation="none")
ax[3].set_title("ASCM ")
for i in range(4):
    ax[i].set_axis_off()

plt.savefig("ascm_comparison.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Comparing outputs of the NLMEANS and ASCM.
#
#
# From the above figure, we can observe that the information of two
# pre-denoised versions of the raw data, ASCM outperforms standard non-local
# means in suppressing noise and preserving feature sharpness.
#
# References
# ----------
#
# .. footbibliography::
#



import os
from os.path import join as pjoin
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.data import DATA_DIR
from dipy.direction.peaks import PeaksAndMetrics
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.utils import create_nifti_header
from dipy.testing import check_for_warnings
from dipy.testing.decorators import set_random_number_generator, use_xvfb
from dipy.tracking.streamline import Streamlines
from dipy.utils.optpkg import optional_package

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

if has_fury:
    from fury import io, window

    from dipy.viz.horizon.app import horizon

skip_it = use_xvfb == "skip"


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator()
def test_horizon_events(rng):
    # using here MNI template affine 2009a
    affine = np.array(
        [
            [1.0, 0.0, 0.0, -98.0],
            [0.0, 1.0, 0.0, -134.0],
            [0.0, 0.0, 1.0, -72.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    data = 255 * rng.random((197, 233, 189))
    vox_size = (1.0, 1.0, 1.0)
    img = np.zeros((197, 233, 189))
    img[0:25, :, :] = 1
    images = [(data, affine, "/test/filename.nii.gz"), (img, affine)]

    peak_dirs = 255 * rng.random((5, 5, 5, 5, 3))
    pam = PeaksAndMetrics()
    pam.peak_dirs = peak_dirs
    pam.affine = affine
    pams = [pam]

    from dipy.segment.tests.test_bundles import setup_module

    setup_module()
    from dipy.segment.tests.test_bundles import f1

    streamlines = f1.copy()
    streamlines._data += np.array([-98.0, -134.0, -72.0])

    header = create_nifti_header(affine, data.shape, vox_size)
    sft = StatefulTractogram(streamlines, header, Space.RASMM)

    tractograms = [sft]

    # select all centroids and expand and click everything else
    # do not press the key shortcuts as vtk generates warning that
    # blocks recording
    fname = os.path.join(DATA_DIR, "record_horizon.log.gz")

    with TemporaryDirectory() as out_dir:
        horizon(
            tractograms=tractograms,
            images=images,
            pams=pams,
            cluster=True,
            cluster_thr=5.0,
            roi_images=True,
            random_colors=False,
            length_gt=0,
            length_lt=np.inf,
            clusters_gt=0,
            clusters_lt=np.inf,
            world_coords=True,
            interactive=True,
            out_png=pjoin(out_dir, "horizon-event.png"),
            recorded_events=fname,
        )


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator()
def test_horizon(rng):
    s1 = 10 * np.array(
        [[0, 0, 0], [1, 0, 0], [2, 0, 0], [3, 0, 0], [4, 0, 0]], dtype="f8"
    )

    s2 = 10 * np.array(
        [[0, 0, 0], [0, 1, 0], [0, 2, 0], [0, 3, 0], [0, 4, 0]], dtype="f8"
    )

    s3 = 10 * np.array(
        [[0, 0, 0], [1, 0.2, 0], [2, 0.2, 0], [3, 0.2, 0], [4, 0.2, 0]], dtype="f8"
    )

    streamlines = Streamlines()
    streamlines.append(s1)
    streamlines.append(s2)
    streamlines.append(s3)
    streamlines.shrink_data()

    affine = np.array(
        [
            [1.0, 0.0, 0.0, -98.0],
            [0.0, 1.0, 0.0, -134.0],
            [0.0, 0.0, 1.0, -72.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    data = 255 * rng.random((197, 233, 189))
    vox_size = (1.0, 1.0, 1.0)

    streamlines._data += np.array([-98.0, -134.0, -72.0])

    header = create_nifti_header(affine, data.shape, vox_size)
    sft = StatefulTractogram(streamlines, header, Space.RASMM)

    # only tractograms
    tractograms = [sft]
    images = None
    with TemporaryDirectory() as out_dir:
        horizon(
            tractograms=tractograms,
            images=images,
            cluster=True,
            cluster_thr=5,
            random_colors=False,
            length_lt=np.inf,
            length_gt=0,
            clusters_lt=np.inf,
            clusters_gt=0,
            world_coords=True,
            interactive=False,
            out_png=pjoin(out_dir, "only-tractograms.png"),
        )

        images = [(data, affine, "/test/filename.nii.gz")]
        # tractograms in native coords (not supported for now)
        with npt.assert_raises(ValueError) as ve:
            horizon(
                tractograms=tractograms,
                images=images,
                cluster=True,
                cluster_thr=5,
                random_colors=False,
                length_lt=np.inf,
                length_gt=0,
                clusters_lt=np.inf,
                clusters_gt=0,
                world_coords=False,
                interactive=False,
                out_png=pjoin(out_dir, "native-tractograms.png"),
            )

        msg = "Currently native coordinates are not supported for streamlines."
        npt.assert_(msg in str(ve.exception))

        # only images
        tractograms = None
        horizon(
            tractograms=tractograms,
            images=images,
            cluster=True,
            cluster_thr=5,
            random_colors=False,
            length_lt=np.inf,
            length_gt=0,
            clusters_lt=np.inf,
            clusters_gt=0,
            world_coords=True,
            interactive=False,
            out_png=pjoin(out_dir, "only-images.png"),
        )

        # no clustering tractograms and images
        horizon(
            tractograms=tractograms,
            images=images,
            cluster=False,
            cluster_thr=5,
            random_colors=False,
            length_lt=np.inf,
            length_gt=0,
            clusters_lt=np.inf,
            clusters_gt=0,
            world_coords=True,
            interactive=False,
            out_png=pjoin(out_dir, "no-clusting-tractograms-and-images.png"),
        )


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_horizon_wrong_dtype_images():
    affine = np.array(
        [
            [1.0, 0.0, 0.0, -98.0],
            [0.0, 1.0, 0.0, -134.0],
            [0.0, 0.0, 1.0, -72.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )

    data = np.random.rand(197, 233, 189).astype(np.bool_)
    images = [(data, affine)]
    with TemporaryDirectory() as out_dir:
        horizon(
            images=images,
            interactive=False,
            out_png=pjoin(out_dir, "wrong-dtype.png"),
        )
        # Asserting the image will not get added and the image will be black.
        assert len(np.unique(io.load_image(pjoin(out_dir, "wrong-dtype.png")))) == 1


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator(42)
def test_roi_images(rng):
    img1 = rng.random((5, 5, 5))
    img2 = np.zeros((5, 5, 5))
    img2[2, 2, 2] = 1
    img3 = np.zeros((5, 5, 5))
    img3[0, :, :] = 1
    images = [
        (img1, np.eye(4)),
        (img2, np.eye(4), "/test/filename.nii.gz"),
        (img3, np.eye(4), "/test/filename.nii.gz"),
    ]
    show_m = horizon(images=images, return_showm=True)
    analysis = window.analyze_scene(show_m.scene)
    npt.assert_equal(analysis.actors, 0)
    arr = window.snapshot(show_m.scene)
    report = window.analyze_snapshot(arr, colors=[(0, 0, 0), (255, 255, 255)])
    npt.assert_array_equal(report.colors_found, [True, True])
    show_m = horizon(images=images, roi_images=True, return_showm=True)
    analysis = window.analyze_scene(show_m.scene)
    npt.assert_equal(analysis.actors, 3)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
@set_random_number_generator(42)
def test_surfaces(rng):
    vertices = rng.random((100, 3))
    faces = rng.integers(0, 100, size=(100, 3))
    surfaces = [
        (vertices, faces),
        (vertices, faces, "/test/filename.pial"),
        (vertices, faces, "/test/filename.pial"),
    ]
    show_m = horizon(surfaces=surfaces, return_showm=True)
    analysis = window.analyze_scene(show_m.scene)
    npt.assert_equal(analysis.actors, 3)

    vertices = rng.random((100, 4))
    faces = rng.integers(0, 100, size=(100, 3))
    surfaces = [
        (vertices, faces),
        (vertices, faces, "/test/filename.pial"),
        (vertices, faces, "/test/filename.pial"),
    ]

    with warnings.catch_warnings(record=True) as l_warns:
        show_m = horizon(surfaces=surfaces, return_showm=True)
        analysis = window.analyze_scene(show_m.scene)
        npt.assert_equal(analysis.actors, 0)
        check_for_warnings(l_warns, "Vertices do not have correct shape: (100, 4)")

    vertices = rng.random((100, 3))
    faces = rng.integers(0, 100, size=(100, 4))
    surfaces = [
        (vertices, faces),
        (vertices, faces, "/test/filename.pial"),
        (vertices, faces, "/test/filename.pial"),
    ]

    with warnings.catch_warnings(record=True) as l_warns:
        show_m = horizon(surfaces=surfaces, return_showm=True)
        analysis = window.analyze_scene(show_m.scene)
        npt.assert_equal(analysis.actors, 0)
        check_for_warnings(l_warns, "Faces do not have correct shape: (100, 4)")


@pytest.mark.skipif(skip_it, reason="Needs xvfb")
def test_small_horizon_import():
    from dipy.viz import horizon as Horizon

    if has_fury:
        assert Horizon == horizon
    else:
        npt.assert_raises(ImportError, Horizon)



import itertools

import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
)

import dipy.segment.featurespeed as dipysfeature
import dipy.segment.metric as dipymetric
import dipy.segment.metricspeed as dipysmetric
from dipy.testing import (
    assert_false,
    assert_greater_equal,
    assert_less_equal,
    assert_true,
)
from dipy.testing.decorators import set_random_number_generator


def norm(x, order=None, axis=None):
    if axis is not None:
        return np.apply_along_axis(np.linalg.norm, axis, x.astype(np.float64), order)

    return np.linalg.norm(x.astype(np.float64), ord=order)


dtype = "float32"

# Create wiggling streamline
nb_points = 18
rng = np.random.default_rng(42)
x = np.linspace(0, 10, nb_points)
y = rng.random(nb_points)
z = np.sin(np.linspace(0, np.pi, nb_points))  # Bending
s = np.array([x, y, z], dtype=dtype).T

# Create trivial streamlines
s1 = np.array([np.arange(10, dtype=dtype)] * 3).T  # 10x3
s2 = np.arange(3 * 10, dtype=dtype).reshape((-1, 3))[::-1]  # 10x3
s3 = np.array([np.arange(5, dtype=dtype)] * 4)  # 5x4
s4 = np.array([np.arange(5, dtype=dtype)] * 3)  # 5x3
streamlines = [s, s1, s2, s3, s4]


def test_metric_minimum_average_direct_flip():
    feature = dipysfeature.IdentityFeature()

    class MinimumAverageDirectFlipMetric(dipysmetric.Metric):
        def __init__(self, feature):
            super(MinimumAverageDirectFlipMetric, self).__init__(feature=feature)

        @property
        def is_order_invariant(self):
            return True  # Ordering is handled in the distance computation

        def are_compatible(self, shape1, shape2):
            return shape1[0] == shape2[0]

        def dist(self, v1, v2):
            def average_euclidean(x, y):
                return np.mean(norm(x - y, axis=1))

            dist_direct = average_euclidean(v1, v2)
            dist_flipped = average_euclidean(v1, v2[::-1])
            return min(dist_direct, dist_flipped)

    for metric in [
        MinimumAverageDirectFlipMetric(feature),
        dipysmetric.MinimumAverageDirectFlipMetric(feature),
    ]:
        # Test special cases of the MDF distance.
        assert_equal(metric.dist(s, s), 0.0)
        assert_equal(metric.dist(s, s[::-1]), 0.0)

        # Translation
        offset = np.array([0.8, 1.3, 5], dtype=dtype)
        assert_almost_equal(metric.dist(s, s + offset), norm(offset), 5)

        # Scaling
        M_scaling = np.diag([1.2, 2.8, 3]).astype(dtype)
        s_mean = np.mean(s, axis=0)
        s_zero_mean = s - s_mean
        s_scaled = np.dot(M_scaling, s_zero_mean.T).T + s_mean
        d = np.mean(norm((np.diag(M_scaling) - 1) * s_zero_mean, axis=1))
        assert_almost_equal(metric.dist(s, s_scaled), d, 5)

        # Rotation
        from dipy.core.geometry import rodrigues_axis_rotation

        rot_axis = np.array([1, 2, 3], dtype=dtype)
        M_rotation = rodrigues_axis_rotation(rot_axis, 60.0).astype(dtype)
        s_mean = np.mean(s, axis=0)
        s_zero_mean = s - s_mean
        s_rotated = np.dot(M_rotation, s_zero_mean.T).T + s_mean

        opposite = norm(np.cross(rot_axis, s_zero_mean), axis=1) / norm(rot_axis)
        distances = np.sqrt(
            2 * opposite**2 * (1 - np.cos(60.0 * np.pi / 180.0))
        ).astype(dtype)
        d = np.mean(distances)
        assert_almost_equal(metric.dist(s, s_rotated), d, 5)

        # All possible pairs
        for s1, s2 in itertools.product(*[streamlines] * 2):
            # Extract features since metric doesn't work
            # directly on streamlines
            f1 = metric.feature.extract(s1)
            f2 = metric.feature.extract(s2)

            # Test method are_compatible
            same_nb_points = f1.shape[0] == f2.shape[0]
            assert_equal(metric.are_compatible(f1.shape, f2.shape), same_nb_points)

            # Test method dist if features are compatible
            if metric.are_compatible(f1.shape, f2.shape):
                distance = metric.dist(f1, f2)
                if np.all(f1 == f2):
                    assert_equal(distance, 0.0)

                assert_almost_equal(distance, dipysmetric.dist(metric, s1, s2))
                assert_almost_equal(distance, dipymetric.mdf(s1, s2))
                assert_greater_equal(distance, 0.0)

        # This metric type is order invariant
        assert_true(metric.is_order_invariant)
        # All possible pairs
        for s1, s2 in itertools.product(*[streamlines] * 2):
            f1 = metric.feature.extract(s1)
            f2 = metric.feature.extract(s2)

            if not metric.are_compatible(f1.shape, f2.shape):
                continue

            f1_flip = metric.feature.extract(s1[::-1])
            f2_flip = metric.feature.extract(s2[::-1])

            distance = metric.dist(f1, f2)
            assert_almost_equal(metric.dist(f1_flip, f2_flip), distance)

            if not np.all(f1_flip == f2_flip):
                assert_true(np.allclose(metric.dist(f1, f2_flip), distance))
                assert_true(np.allclose(metric.dist(f1_flip, f2), distance))


def test_metric_cosine():
    feature = dipysfeature.VectorOfEndpointsFeature()

    class CosineMetric(dipysmetric.Metric):
        def __init__(self, feature):
            super(CosineMetric, self).__init__(feature=feature)

        def are_compatible(self, shape1, shape2):
            # Cosine metric works on vectors.
            return shape1 == shape2 and shape1[0] == 1

        def dist(self, v1, v2):
            # Check if we have null vectors
            if norm(v1) == 0:
                return 0.0 if norm(v2) == 0 else 1.0

            v1_normed = v1.astype(np.float64) / norm(v1.astype(np.float64))
            v2_normed = v2.astype(np.float64) / norm(v2.astype(np.float64))
            cos_theta = np.dot(v1_normed, v2_normed.T).item()
            # Make sure it's in [-1, 1], i.e. within domain of arccosine
            cos_theta = np.minimum(cos_theta, 1.0)
            cos_theta = np.maximum(cos_theta, -1.0)
            return np.arccos(cos_theta) / np.pi  # Normalized cosine distance

    for metric in [CosineMetric(feature), dipysmetric.CosineMetric(feature)]:
        # Test special cases of the cosine distance.
        v0 = np.array([[0, 0, 0]], dtype=np.float32)
        v1 = np.array([[1, 2, 3]], dtype=np.float32)
        v2 = np.array([[1, -1.0 / 2, 0]], dtype=np.float32)
        v3 = np.array([[-1, -2, -3]], dtype=np.float32)

        assert_equal(metric.dist(v0, v0), 0.0)  # dot-dot
        assert_equal(metric.dist(v0, v1), 1.0)  # dot-line
        assert_equal(metric.dist(v1, v1), 0.0)  # collinear
        assert_equal(metric.dist(v1, v2), 0.5)  # orthogonal
        assert_equal(metric.dist(v1, v3), 1.0)  # opposite

        # All possible pairs
        for s1, s2 in itertools.product(*[streamlines] * 2):
            # Extract features since metric doesn't
            # work directly on streamlines
            f1 = metric.feature.extract(s1)
            f2 = metric.feature.extract(s2)

            # Test method are_compatible
            are_vectors = f1.shape[0] == 1 and f2.shape[0] == 1
            same_dimension = f1.shape[1] == f2.shape[1]
            assert_equal(
                metric.are_compatible(f1.shape, f2.shape),
                are_vectors and same_dimension,
            )

            # Test method dist if features are compatible
            if metric.are_compatible(f1.shape, f2.shape):
                distance = metric.dist(f1, f2)
                if np.all(f1 == f2):
                    assert_almost_equal(distance, 0.0)

                assert_almost_equal(distance, dipysmetric.dist(metric, s1, s2))
                assert_greater_equal(distance, 0.0)
                assert_less_equal(distance, 1.0)

        # This metric type is not order invariant
        assert_false(metric.is_order_invariant)
        # All possible pairs
        for s1, s2 in itertools.product(*[streamlines] * 2):
            f1 = metric.feature.extract(s1)
            f2 = metric.feature.extract(s2)

            if not metric.are_compatible(f1.shape, f2.shape):
                continue

            f1_flip = metric.feature.extract(s1[::-1])
            f2_flip = metric.feature.extract(s2[::-1])

            distance = metric.dist(f1, f2)
            assert_almost_equal(metric.dist(f1_flip, f2_flip), distance)

            if not np.all(f1_flip == f2_flip):
                assert_false(metric.dist(f1, f2_flip) == distance)
                assert_false(metric.dist(f1_flip, f2) == distance)


def test_subclassing_metric():
    class EmptyMetric(dipysmetric.Metric):
        pass

    metric = EmptyMetric()
    assert_raises(NotImplementedError, metric.are_compatible, None, None)
    assert_raises(NotImplementedError, metric.dist, None, None)


@set_random_number_generator()
def test_distance_matrix(rng):
    metric = dipysmetric.SumPointwiseEuclideanMetric()

    for dtype in [np.int32, np.int64, np.float32, np.float64]:
        # Compute distances of all tuples spawn by the Cartesian product
        # of `data` with itself.
        data = (rng.random((4, 10, 3)) * 10).astype(dtype)
        D = dipysmetric.distance_matrix(metric, data)
        assert_equal(D.shape, (len(data), len(data)))
        assert_array_equal(np.diag(D), np.zeros(len(data)))

        if metric.is_order_invariant:
            # Distance matrix should be symmetric
            assert_array_equal(D, D.T)

        for i in range(len(data)):
            for j in range(len(data)):
                assert_equal(D[i, j], dipysmetric.dist(metric, data[i], data[j]))

        # Compute distances of all tuples spawn by the Cartesian product
        # of `data` with `data2`.
        data2 = (rng.random((3, 10, 3)) * 10).astype(dtype)
        D = dipysmetric.distance_matrix(metric, data, data2)
        assert_equal(D.shape, (len(data), len(data2)))

        for i in range(len(data)):
            for j in range(len(data2)):
                assert_equal(D[i, j], dipysmetric.dist(metric, data[i], data2[j]))


@set_random_number_generator()
def test_mean_distances(rng):
    nb_slines = 10
    nb_pts = 22
    dim = 3
    a = rng.random((nb_slines, nb_pts, dim))
    b = rng.random((nb_slines, nb_pts, dim))
    diff = a - b

    # Test Euclidean distance (L2)
    mean_l2_dist = dipymetric.mean_euclidean_distance(a, b)
    diff_norm = np.linalg.norm(diff.reshape((-1, dim)), ord=2, axis=-1)
    mean_norm = np.mean(diff_norm.reshape((nb_slines, -1)), axis=-1)
    assert_almost_equal(mean_l2_dist, mean_norm)

    # Test Manhattan distance (L1)
    mean_l1_dist = dipymetric.mean_manhattan_distance(a, b)
    diff_norm = np.linalg.norm(diff.reshape((-1, dim)), ord=1, axis=-1)
    mean_norm = np.mean(diff_norm.reshape((nb_slines, -1)), axis=-1)
    assert_almost_equal(mean_l1_dist, mean_norm)



from copy import deepcopy
import logging
import os
import time

import nibabel as nib
from nibabel.streamlines import detect_format
from nibabel.streamlines.tractogram import Tractogram
import numpy as np
import trx.trx_file_memmap as tmm

from dipy.io.dpy import Dpy
from dipy.io.stateful_tractogram import Origin, Space, StatefulTractogram
from dipy.io.utils import create_tractogram_header, is_header_compatible
from dipy.io.vtk import load_vtk_streamlines, save_vtk_streamlines
from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def save_tractogram(sft, filename, *, bbox_valid_check=True):
    """Save the stateful tractogram in any format (trx/trk/tck/vtk/vtp/fib/dpy)

    Parameters
    ----------
    sft : StatefulTractogram
        The stateful tractogram to save
    filename : string
        Filename with valid extension
    bbox_valid_check : bool
        Verification for negative voxel coordinates or values above the
        volume dimensions. Default is True, to enforce valid file.

    Returns
    -------
    output : bool
        True if the saving operation was successful
    """

    _, extension = os.path.splitext(filename)
    if extension not in [".trk", ".tck", ".trx", ".vtk", ".vtp", ".fib", ".dpy"]:
        raise TypeError("Output filename is not one of the supported format.")

    if bbox_valid_check and not sft.is_bbox_in_vox_valid():
        raise ValueError(
            "Bounding box is not valid in voxel space, cannot "
            "load a valid file if some coordinates are invalid.\n"
            "Please set bbox_valid_check to False and then use "
            "the function remove_invalid_streamlines to discard "
            "invalid streamlines."
        )

    old_space = deepcopy(sft.space)
    old_origin = deepcopy(sft.origin)

    sft.to_rasmm()
    sft.to_center()

    timer = time.time()
    if extension in [".trk", ".tck"]:
        tractogram_type = detect_format(filename)
        header = create_tractogram_header(tractogram_type, *sft.space_attributes)
        new_tractogram = Tractogram(sft.streamlines, affine_to_rasmm=np.eye(4))

        if extension == ".trk":
            new_tractogram.data_per_point = sft.data_per_point
            new_tractogram.data_per_streamline = sft.data_per_streamline

        fileobj = tractogram_type(new_tractogram, header=header)
        nib.streamlines.save(fileobj, filename)

    elif extension in [".vtk", ".vtp", ".fib"]:
        binary = extension in [".vtk", ".fib"]
        save_vtk_streamlines(sft.streamlines, filename, binary=binary)
    elif extension in [".dpy"]:
        dpy_obj = Dpy(filename, mode="w")
        dpy_obj.write_tracks(sft.streamlines)
        dpy_obj.close()
    elif extension in [".trx"]:
        trx = tmm.TrxFile.from_sft(sft)
        tmm.save(trx, filename)
        trx.close()

    logging.debug(
        "Save %s with %s streamlines in %s seconds.",
        filename,
        len(sft),
        round(time.time() - timer, 3),
    )

    sft.to_space(old_space)
    sft.to_origin(old_origin)

    return True


@warning_for_keywords()
def load_tractogram(
    filename,
    reference,
    *,
    to_space=Space.RASMM,
    to_origin=Origin.NIFTI,
    bbox_valid_check=True,
    trk_header_check=True,
):
    """Load the stateful tractogram from any format (trx/trk/tck/vtk/vtp/fib/dpy)

    Parameters
    ----------
    filename : string
        Filename with valid extension
    reference : Nifti or Trk filename, Nifti1Image or TrkFile, Nifti1Header or
        trk.header (dict), or 'same' if the input is a trk file.
        Reference that provides the spatial attribute.
        Typically a nifti-related object from the native diffusion used for
        streamlines generation
    to_space : Enum (dipy.io.stateful_tractogram.Space)
        Space to which the streamlines will be transformed after loading
    to_origin : Enum (dipy.io.stateful_tractogram.Origin)
        Origin to which the streamlines will be transformed after loading
            NIFTI standard, default (center of the voxel)
            TRACKVIS standard (corner of the voxel)
    bbox_valid_check : bool
        Verification for negative voxel coordinates or values above the
        volume dimensions. Default is True, to enforce valid file.
    trk_header_check : bool
        Verification that the reference has the same header as the spatial
        attributes as the input tractogram when a Trk is loaded

    Returns
    -------
    output : StatefulTractogram
        The tractogram to load (must have been saved properly)
    """
    _, extension = os.path.splitext(filename)
    if extension not in [".trk", ".tck", ".trx", ".vtk", ".vtp", ".fib", ".dpy"]:
        logging.error("Output filename is not one of the supported format.")
        return False

    if to_space not in Space:
        logging.error("Space MUST be one of the 3 choices (Enum).")
        return False

    if reference == "same":
        if extension in [".trk", ".trx"]:
            reference = filename
        else:
            logging.error(
                'Reference must be provided, "same" is only ' "available for Trk file."
            )
            return False

    if trk_header_check and extension == ".trk":
        if not is_header_compatible(filename, reference):
            logging.error("Trk file header does not match the provided reference.")
            return False

    timer = time.time()
    data_per_point = None
    data_per_streamline = None
    if extension in [".trk", ".tck"]:
        tractogram_obj = nib.streamlines.load(filename).tractogram
        streamlines = tractogram_obj.streamlines
        if extension == ".trk":
            data_per_point = tractogram_obj.data_per_point
            data_per_streamline = tractogram_obj.data_per_streamline

    elif extension in [".vtk", ".vtp", ".fib"]:
        streamlines = load_vtk_streamlines(filename)
    elif extension in [".dpy"]:
        dpy_obj = Dpy(filename, mode="r")
        streamlines = list(dpy_obj.read_tracks())
        dpy_obj.close()

    if extension in [".trx"]:
        trx_obj = tmm.load(filename)
        sft = trx_obj.to_sft()
        trx_obj.close()
    else:
        sft = StatefulTractogram(
            streamlines,
            reference,
            Space.RASMM,
            origin=Origin.NIFTI,
            data_per_point=data_per_point,
            data_per_streamline=data_per_streamline,
        )

    logging.debug(
        "Load %s with %s streamlines in %s seconds.",
        filename,
        len(sft),
        round(time.time() - timer, 3),
    )

    if bbox_valid_check and not sft.is_bbox_in_vox_valid():
        raise ValueError(
            "Bounding box is not valid in voxel space, cannot "
            "load a valid file if some coordinates are invalid.\n"
            "Please set bbox_valid_check to False and then use "
            "the function remove_invalid_streamlines to discard "
            "invalid streamlines."
        )

    sft.to_space(to_space)
    sft.to_origin(to_origin)

    return sft


def load_generator(ttype):
    """Generate a loading function that performs a file extension
    check to restrict the user to a single file format.

    Parameters
    ----------
    ttype : string
        Extension of the file format that requires a loader

    Returns
    -------
    output : function
        Function (load_tractogram) that handle only one file format
    """

    @warning_for_keywords()
    def f_gen(
        filename,
        reference,
        *,
        to_space=Space.RASMM,
        to_origin=Origin.NIFTI,
        bbox_valid_check=True,
        trk_header_check=True,
    ):
        _, extension = os.path.splitext(filename)
        if not extension == ttype:
            msg = f"This function can only load {ttype} files, "
            msg += "for a more general purpose, use load_tractogram instead."
            raise ValueError(msg)

        sft = load_tractogram(
            filename,
            reference,
            to_space=Space.RASMM,
            to_origin=to_origin,
            bbox_valid_check=bbox_valid_check,
            trk_header_check=trk_header_check,
        )
        return sft

    f_gen.__doc__ = load_tractogram.__doc__.replace(
        "from any format (trk/tck/vtk/vtp/fib/dpy)", f"of the {ttype} format"
    )
    return f_gen


def save_generator(ttype):
    """Generate a saving function that performs a file extension
    check to restrict the user to a single file format.

    Parameters
    ----------
    ttype : string
        Extension of the file format that requires a saver

    Returns
    -------
    output : function
        Function (save_tractogram) that handle only one file format
    """

    def f_gen(sft, filename, bbox_valid_check=True):
        _, extension = os.path.splitext(filename)
        if not extension == ttype:
            msg = f"This function can only save {ttype} file, "
            msg += "for more general cases, use save_tractogram instead."
            raise ValueError(msg)
        save_tractogram(sft, filename, bbox_valid_check=bbox_valid_check)

    f_gen.__doc__ = save_tractogram.__doc__.replace(
        "in any format (trk/tck/vtk/vtp/fib/dpy)", f"of the {ttype} format"
    )
    return f_gen


load_trk = load_generator(".trk")
load_tck = load_generator(".tck")
load_trx = load_generator(".trx")
load_vtk = load_generator(".vtk")
load_vtp = load_generator(".vtp")
load_fib = load_generator(".fib")
load_dpy = load_generator(".dpy")
save_trk = save_generator(".trk")
save_tck = save_generator(".tck")
save_trx = save_generator(".trx")
save_vtk = save_generator(".vtk")
save_vtp = save_generator(".vtp")
save_fib = save_generator(".fib")
save_dpy = save_generator(".dpy")



# init for io routines

from . import utils
from .dpy import Dpy
from .gradients import read_bvals_bvecs
from .pickles import load_pickle, save_pickle

__all__ = ["read_bvals_bvecs", "Dpy", "save_pickle", "load_pickle", "utils"]



import sys

import pytest
import trx.trx_file_memmap as tmm

from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.utils.tractogram import concatenate_tractogram

is_big_endian = "big" in sys.byteorder.lower()


@pytest.mark.skipif(is_big_endian, reason="Little Endian architecture required")
def test_concatenate():
    filepath_dix, _, _ = get_fnames(name="gold_standard_tracks")
    sft = load_tractogram(filepath_dix["gs.trk"], filepath_dix["gs.nii"])
    trx = tmm.load(filepath_dix["gs.trx"])
    concat = concatenate_tractogram([sft, trx])

    assert len(concat) == 2 * len(trx)
    trx.close()
    concat.close()



"""
=================================
Linear fascicle evaluation (LiFE)
=================================

Evaluating the results of tractography algorithms is one of the biggest
challenges for diffusion MRI. One proposal for evaluation of tractography
results is to use a forward model that predicts the signal from each of a set
of streamlines, and then fit a linear model to these simultaneous predictions
:footcite:p:`Pestilli2014`.

We will use streamlines generated using probabilistic tracking on CSA
peaks. For brevity, we will include in this example only streamlines going
through the corpus callosum connecting left to right superior frontal
cortex. The process of tracking and finding these streamlines is fully
demonstrated in the
:ref:`sphx_glr_examples_built_streamline_analysis_streamline_tools.py` example.
If this example has been run, we can read the streamlines from file. Otherwise,
we'll run that example first, by importing it. This provides us with all of the
variables that were created in that example:

"""

from os.path import join as pjoin

import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import AxesGrid
import numpy as np

# We'll need to know where the corpus callosum is from these variables:
from dipy.core.gradients import gradient_table
import dipy.core.optimize as opt
from dipy.data import fetch_stanford_tracks, get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.streamline import load_trk
import dipy.tracking.life as life
from dipy.viz import actor, colormap as cmap, window

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")
t1_fname = get_fnames(name="stanford_t1")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
t1_data = load_nifti_data(t1_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

cc_slice = labels == 2

# Let's now fetch a set of streamlines from the Stanford HARDI dataset.
# Those streamlines were generated during the
# :ref:`sphx_glr_examples_built_streamline_analysis_streamline_tools.py` example.
# Read the candidates from file in voxel space:

streamlines_files = fetch_stanford_tracks()
lr_superiorfrontal_path = pjoin(streamlines_files[1], "hardi-lr-superiorfrontal.trk")

candidate_sl_sft = load_trk(lr_superiorfrontal_path, "same")
candidate_sl_sft.to_vox()
candidate_sl = candidate_sl_sft.streamlines

###############################################################################
# The streamlines that are entered into the model are termed 'candidate
# streamlines' (or a 'candidate connectome'):
#
#
# Let's visualize the initial candidate group of streamlines in 3D, relative
# to the anatomical structure of this brain:

# Enables/disables interactive visualization
interactive = False

candidate_streamlines_actor = actor.streamtube(
    candidate_sl, colors=cmap.line_colors(candidate_sl)
)
cc_ROI_actor = actor.contour_from_roi(cc_slice, color=(1.0, 1.0, 0.0), opacity=0.5)

vol_actor = actor.slicer(t1_data)

vol_actor.display(x=40)
vol_actor2 = vol_actor.copy()
vol_actor2.display(z=35)

# Add display objects to canvas
scene = window.Scene()
scene.add(candidate_streamlines_actor)
scene.add(cc_ROI_actor)
scene.add(vol_actor)
scene.add(vol_actor2)
window.record(scene=scene, n_frames=1, out_path="life_candidates.png", size=(800, 800))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Candidate connectome before life optimization
#
#
#
# Next, we initialize a LiFE model. We import the ``dipy.tracking.life``
# module, which contains the classes and functions that implement the model:

fiber_model = life.FiberModel(gtab)

###############################################################################
# Since we read the streamlines from a file, already in the voxel space, we do
# not need to transform them into this space. Otherwise, if the streamline
# coordinates were in the world space (relative to the scanner iso-center, or
# relative to the mid-point of the AC-PC-connecting line), we would use this::
#
#    inv_affine = np.linalg.inv(hardi_img.affine)
#
# the inverse transformation from world space to the voxel space as the affine
# for the following model fit.
#
# The next step is to fit the model, producing a ``FiberFit`` class instance,
# that stores the data, as well as the results of the fitting procedure.
#
# The LiFE model posits that the signal in the diffusion MRI volume can be
# explained by the streamlines, by the equation
#
# .. math::
#
#     y = X\beta
#
# Where $y$ is the diffusion MRI signal, $\beta$ are a set of weights on the
# streamlines and $X$ is a design matrix. This matrix has the dimensions $m$ by
# $n$, where $m=n_{voxels} \cdot n_{directions}$, and $n_{voxels}$ is the set
# of voxels in the ROI that contains the streamlines considered in this model.
# The $i^{th}$ column of the matrix contains the expected contributions of the
# $i^{th}$ streamline (arbitrarily ordered) to each of the voxels. $X$ is a
# sparse matrix, because each streamline traverses only a small percentage of
# the voxels. The  expected contributions of the streamline are calculated
# using a forward model, where each node of the streamline is modeled as a
# cylindrical fiber compartment with Gaussian diffusion, using the diffusion
# tensor model. See :footcite:p:`Pestilli2014` for more detail on the model, and
# variations of this model.

fiber_fit = fiber_model.fit(data, candidate_sl, affine=np.eye(4))

###############################################################################
# The ``FiberFit`` class instance holds various properties of the model fit.
# For example, it has the weights $\beta$, that are assigned to each
# streamline. In most cases, a tractography through some region will include
# redundant streamlines, and these streamlines will have $\beta_i$ that are 0.

fig, ax = plt.subplots(1)
ax.hist(fiber_fit.beta, bins=100, histtype="step")
ax.set_xlabel("Fiber weights")
ax.set_ylabel("# fibers")
fig.savefig("beta_histogram.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# LiFE streamline weights
#
#
#
# We use $\beta$ to filter out these redundant streamlines, and generate an
# optimized group of streamlines:

optimized_sl = [np.vstack(candidate_sl)[np.where(fiber_fit.beta > 0)[0]]]
scene = window.Scene()
scene.add(actor.streamtube(optimized_sl, colors=cmap.line_colors(optimized_sl)))
scene.add(cc_ROI_actor)
scene.add(vol_actor)
window.record(scene=scene, n_frames=1, out_path="life_optimized.png", size=(800, 800))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Streamlines selected via LiFE optimization
#
#
#
#
# The new set of streamlines should do well in fitting the data, and redundant
# streamlines have presumably been removed (in this case, about 50% of the
# streamlines).
#
# But how well does the model do in explaining the diffusion data? We can
# quantify that: the ``FiberFit`` class instance has a `predict` method, which
# can be used to invert the model and predict back either the data that was
# used to fit the model, or other unseen data (e.g. in cross-validation, see
# :ref:`sphx_glr_examples_built_reconstruction_kfold_xval.py`).
#
# Without arguments, the ``.predict()`` method will predict the diffusion
# signal for the same gradient table that was used in the fit data, but
# ``gtab`` and ``S0`` keyword arguments can be used to predict for other
# acquisition schemes and other baseline non-diffusion-weighted signals.

model_predict = fiber_fit.predict()

###############################################################################
# We will focus on the error in prediction of the diffusion-weighted data, and
# calculate the root of the mean squared error.

model_error = model_predict - fiber_fit.data
model_rmse = np.sqrt(np.mean(model_error[:, 10:] ** 2, -1))

###############################################################################
# As a baseline against which we can compare, we calculate another error term.
# In this case, we assume that the weight for each streamline is equal
# to zero. This produces the naive prediction of the mean of the signal in each
# voxel.

beta_baseline = np.zeros(fiber_fit.beta.shape[0])
pred_weighted = np.reshape(
    opt.spdot(fiber_fit.life_matrix, beta_baseline),
    (fiber_fit.vox_coords.shape[0], np.sum(~gtab.b0s_mask)),
)
mean_pred = np.empty((fiber_fit.vox_coords.shape[0], gtab.bvals.shape[0]))
S0 = fiber_fit.b0_signal

###############################################################################
# Since the fitting is done in the demeaned S/S0 domain, we need
# to add back the mean and then multiply by S0 in every voxel:

mean_pred[..., gtab.b0s_mask] = S0[:, None]
mean_pred[..., ~gtab.b0s_mask] = (pred_weighted + fiber_fit.mean_signal[:, None]) * S0[
    :, None
]
mean_error = mean_pred - fiber_fit.data
mean_rmse = np.sqrt(np.mean(mean_error**2, -1))

###############################################################################
# First, we can compare the overall distribution of errors between these two
# alternative models of the ROI. We show the distribution of differences in
# error (improvement through model fitting, relative to the baseline model).
# Here, positive values denote an improvement in error with model fit, relative
# to without the model fit.

fig, ax = plt.subplots(1)
ax.hist(mean_rmse - model_rmse, bins=100, histtype="step")
ax.text(
    0.2,
    0.9,
    f"Median RMSE, mean model: {np.median(mean_rmse):.2f}",
    horizontalalignment="left",
    verticalalignment="center",
    transform=ax.transAxes,
)
ax.text(
    0.2,
    0.8,
    f"Median RMSE, LiFE: {np.median(model_rmse):.2f}",
    horizontalalignment="left",
    verticalalignment="center",
    transform=ax.transAxes,
)
ax.set_xlabel("RMS Error")
ax.set_ylabel("# voxels")
fig.savefig("error_histograms.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Improvement in error with fitting of the LiFE model.
#
#
#
#
#
# Second, we can show the spatial distribution of the two error terms,
# and of the improvement with the model fit:

vol_model = np.ones(data.shape[:3]) * np.nan
vol_model[
    fiber_fit.vox_coords[:, 0], fiber_fit.vox_coords[:, 1], fiber_fit.vox_coords[:, 2]
] = model_rmse
vol_mean = np.ones(data.shape[:3]) * np.nan
vol_mean[
    fiber_fit.vox_coords[:, 0], fiber_fit.vox_coords[:, 1], fiber_fit.vox_coords[:, 2]
] = mean_rmse
vol_improve = np.ones(data.shape[:3]) * np.nan
vol_improve[
    fiber_fit.vox_coords[:, 0], fiber_fit.vox_coords[:, 1], fiber_fit.vox_coords[:, 2]
] = mean_rmse - model_rmse
sl_idx = 49
fig = plt.figure()
fig.subplots_adjust(left=0.05, right=0.95)
ax = AxesGrid(
    fig,
    111,
    nrows_ncols=(1, 3),
    label_mode="1",
    share_all=True,
    cbar_location="top",
    cbar_mode="each",
    cbar_size="10%",
    cbar_pad="5%",
)
ax[0].matshow(np.rot90(t1_data[sl_idx, :, :]), cmap=matplotlib.cm.bone)
im = ax[0].matshow(np.rot90(vol_model[sl_idx, :, :]), cmap=matplotlib.cm.hot)
ax.cbar_axes[0].colorbar(im)
ax[1].matshow(np.rot90(t1_data[sl_idx, :, :]), cmap=matplotlib.cm.bone)
im = ax[1].matshow(np.rot90(vol_mean[sl_idx, :, :]), cmap=matplotlib.cm.hot)
ax.cbar_axes[1].colorbar(im)
ax[2].matshow(np.rot90(t1_data[sl_idx, :, :]), cmap=matplotlib.cm.bone)
im = ax[2].matshow(np.rot90(vol_improve[sl_idx, :, :]), cmap=matplotlib.cm.RdBu)
ax.cbar_axes[2].colorbar(im)
for lax in ax:
    lax.set_xticks([])
    lax.set_yticks([])
fig.savefig("spatial_errors.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Spatial distribution of error and improvement.
#
#
#
#
#
# This image demonstrates that in many places, fitting the LiFE model results
# in substantial reduction of the error.
#
# Note that for full-brain tractographies *LiFE* can require large amounts of
# memory. For detailed memory profiling of the algorithm, based on the
# streamlines generated in
# :ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py`, see
# `this IPython notebook
# <http://nbviewer.ipython.org/gist/arokem/bc29f34ebc97510d9def>`_.
#
# For the Matlab implementation of LiFE, head over to `Franco Pestilli's github
# webpage <https://francopestilli.github.io/life/>`_.
#
# References
# ----------
#
# .. footbibliography::
#



from dipy.testing.decorators import warning_for_keywords
from dipy.workflows.workflow import Workflow


class CombinedWorkflow(Workflow):
    @warning_for_keywords()
    def __init__(
        self, *, output_strategy="append", mix_names=False, force=False, skip=False
    ):
        """Workflow that combines multiple workflows.
        The workflow combined together are referred as sub flows in this class.
        """

        self._optionals = {}
        super(CombinedWorkflow, self).__init__(
            output_strategy=output_strategy, mix_names=mix_names, force=force, skip=skip
        )

    def get_sub_runs(self):
        """Returns a list of tuples
        (sub flow name, sub flow run method, sub flow short name)
        to be used in the sub flow parameters extraction.
        """
        sub_runs = []
        for flow in self._get_sub_flows():
            sub_runs.append((flow.__name__, flow.run, flow.get_short_name()))

        return sub_runs

    def _get_sub_flows(self):
        """Returns a list of sub flows used in the combined_workflow. Needs to
        be implemented in every new combined_workflow.
        """
        raise AttributeError(
            f"Error: _get_sub_flows() has to be defined for {self.__class__}"
        )

    def set_sub_flows_optionals(self, opts):
        """Sets the self._optionals variable with all sub flow arguments
        that were passed in the commandline.
        """
        self._optionals = {}
        for key, sub_dict in opts.items():
            self._optionals[key] = {k: v for k, v in sub_dict.items() if v is not None}

    def get_optionals(self, flow, **kwargs):
        """Returns the sub flow's optional arguments merged with those passed
        as params in kwargs.
        """
        opts = self._optionals[flow.__name__]
        opts.update(kwargs)

        return opts

    def run_sub_flow(self, flow, *args, **kwargs):
        """Runs the sub flow with the optional parameters passed via the
        command line. This is a convenience method to make sub flow running
        more intuitive on the concrete CombinedWorkflow side.
        """
        return flow.run(*args, **self.get_optionals(type(flow), **kwargs))



import warnings

import numpy as np

from dipy.nn.utils import normalize, recover_img, transform_img, unnormalize
from dipy.testing.decorators import set_random_number_generator


@set_random_number_generator()
def test_norm(rng=None):
    temp = rng.random((8, 8, 8)) * 10
    temp2 = normalize(temp)
    temp2 = unnormalize(temp2, -1, 1, 0, 10)
    np.testing.assert_almost_equal(temp, temp2, 1)


@set_random_number_generator()
def test_transform(rng=None):
    temp = rng.random((30, 31, 32))
    temp2, affine, mid_shape, offset_array, scale, crop_vs, pad_vs = transform_img(
        temp, np.eye(4), init_shape=(32, 32, 32), voxsize=np.ones(3) * 2
    )
    with warnings.catch_warnings():
        scipy_affine_txfm_msg = (
            "The behavior of affine_transform with a 1-D "
            "array supplied for the matrix parameter has changed in "
            "SciPy 0.18.0."
        )
        warnings.filterwarnings(
            "ignore", message=scipy_affine_txfm_msg, category=UserWarning
        )
        temp2 = recover_img(
            temp2,
            affine,
            mid_shape,
            temp.shape,
            offset_array,
            np.ones(3) * 2,
            scale,
            crop_vs,
            pad_vs,
        )
    np.testing.assert_almost_equal(np.array(temp.shape), np.array(temp2.shape))



#!/usr/bin/env python3
""" Checkout gitwash repo into directory and do search replace on name """

import fnmatch
import glob
from optparse import OptionParser
import os
from os.path import join as pjoin
import re
import shutil
from subprocess import call
import sys
import tempfile

verbose = False


def clone_repo(url, branch):
    cwd = os.getcwd()
    tmpdir = tempfile.mkdtemp()
    try:
        cmd = f'git clone {url} {tmpdir}'
        call(cmd, shell=True)
        os.chdir(tmpdir)
        cmd = f'git checkout {branch}'
        call(cmd, shell=True)
    except:
        shutil.rmtree(tmpdir)
        raise
    finally:
        os.chdir(cwd)
    return tmpdir


def cp_files(in_path, globs, out_path):
    try:
        os.makedirs(out_path)
    except OSError:
        pass
    out_fnames = []
    for in_glob in globs:
        in_glob_path = pjoin(in_path, in_glob)
        for in_fname in glob.glob(in_glob_path):
            out_fname = in_fname.replace(in_path, out_path)
            pth, _ = os.path.split(out_fname)
            if not os.path.isdir(pth):
                os.makedirs(pth)
            shutil.copyfile(in_fname, out_fname)
            out_fnames.append(out_fname)
    return out_fnames


def filename_search_replace(sr_pairs, filename, backup=False):
    """ Search and replace for expressions in files

    """
    in_txt = open(filename, 'rt').read(-1)
    out_txt = in_txt[:]
    for in_exp, out_exp in sr_pairs:
        in_exp = re.compile(in_exp)
        out_txt = in_exp.sub(out_exp, out_txt)
    if in_txt == out_txt:
        return False
    open(filename, 'wt').write(out_txt)
    if backup:
        open(filename + '.bak', 'wt').write(in_txt)
    return True


def copy_replace(replace_pairs,
                 repo_path,
                 out_path,
                 cp_globs=('*',),
                 rep_globs=('*',),
                 renames=()):
    out_fnames = cp_files(repo_path, cp_globs, out_path)
    renames = [(re.compile(in_exp), out_exp) for in_exp, out_exp in renames]
    fnames = []
    for rep_glob in rep_globs:
        fnames += fnmatch.filter(out_fnames, rep_glob)
    if verbose:
        print('\n'.join(fnames))
    for fname in fnames:
        filename_search_replace(replace_pairs, fname, False)
        for in_exp, out_exp in renames:
            new_fname, n = in_exp.subn(out_exp, fname)
            if n:
                os.rename(fname, new_fname)
                break


def make_link_targets(proj_name,
                      user_name,
                      repo_name,
                      known_link_fname,
                      out_link_fname,
                      url=None,
                      ml_url=None):
    """ Check and make link targets

    If url is None or ml_url is None, check if there are links present for these
    in `known_link_fname`.  If not, raise error.  The check is:

    Look for a target `proj_name`.
    Look for a target `proj_name` + ' mailing list'

    Also, look for a target `proj_name` + 'github'.  If this exists, don't write
    this target into the new file below.

    If we are writing any of the url, ml_url, or github address, then write new
    file with these links, of form:

    .. _`proj_name`
    .. _`proj_name`: url
    .. _`proj_name` mailing list: url
    """
    link_contents = open(known_link_fname, 'rt').readlines()
    have_url = url is not None
    have_ml_url = ml_url is not None
    have_gh_url = None
    for line in link_contents:
        if not have_url:
            match = re.match(r'..\s+_`%s`:\s+' % proj_name, line)
            if match:
                have_url = True
        if not have_ml_url:
            match = re.match(r'..\s+_`%s mailing list`:\s+' % proj_name, line)
            if match:
                have_ml_url = True
        if not have_gh_url:
            match = re.match(r'..\s+_`%s github`:\s+' % proj_name, line)
            if match:
                have_gh_url = True
    if not have_url or not have_ml_url:
        raise RuntimeError('Need command line or known project '
                           'and / or mailing list URLs')
    lines = []
    if url is not None:
        lines.append(f'.. _`{proj_name}`: {url}\n')
    if not have_gh_url:
        gh_url = f'http://github.com/{user_name}/{repo_name}\n'
        lines.append(f'.. _`{proj_name} github`: {gh_url}\n')
    if ml_url is not None:
        lines.append(f'.. _`{proj_name} mailing list`: {ml_url}\n')
    if len(lines) == 0:
        # Nothing to do
        return
    # A neat little header line
    lines = [f'.. {proj_name}\n'] + lines
    out_links = open(out_link_fname, 'wt')
    out_links.writelines(lines)
    out_links.close()


USAGE = """ <output_directory> <project_name>

If not set with options, the repository name is the same as the <project
name>

If not set with options, the main github user is the same as the
repository name."""


GITWASH_CENTRAL = 'git://github.com/matthew-brett/gitwash.git'
GITWASH_BRANCH = 'master'


def main():
    parser = OptionParser()
    parser.set_usage(parser.get_usage().strip() + USAGE)
    parser.add_option("--repo-name", dest="repo_name",
                      help="repository name - e.g. nitime",
                      metavar="REPO_NAME")
    parser.add_option("--github-user", dest="main_gh_user",
                      help="github username for main repo - e.g fperez",
                      metavar="MAIN_GH_USER")
    parser.add_option("--gitwash-url", dest="gitwash_url",
                      help=f"URL to gitwash repository - default {GITWASH_CENTRAL}",
                      default=GITWASH_CENTRAL,
                      metavar="GITWASH_URL")
    parser.add_option("--gitwash-branch", dest="gitwash_branch",
                      help=f"branch in gitwash repository - default {GITWASH_BRANCH}",
                      default=GITWASH_BRANCH,
                      metavar="GITWASH_BRANCH")
    parser.add_option("--source-suffix", dest="source_suffix",
                      help="suffix of ReST source files - default '.rst'",
                      default='.rst',
                      metavar="SOURCE_SUFFIX")
    parser.add_option("--project-url", dest="project_url",
                      help="URL for project web pages",
                      default=None,
                      metavar="PROJECT_URL")
    parser.add_option("--project-ml-url", dest="project_ml_url",
                      help="URL for project mailing list",
                      default=None,
                      metavar="PROJECT_ML_URL")
    (options, args) = parser.parse_args()
    if len(args) < 2:
        parser.print_help()
        sys.exit()
    out_path, project_name = args
    if options.repo_name is None:
        options.repo_name = project_name
    if options.main_gh_user is None:
        options.main_gh_user = options.repo_name
    repo_path = clone_repo(options.gitwash_url, options.gitwash_branch)
    try:
        copy_replace((('PROJECTNAME', project_name),
                      ('REPONAME', options.repo_name),
                      ('MAIN_GH_USER', options.main_gh_user)),
                     repo_path,
                     out_path,
                     cp_globs=(pjoin('gitwash', '*'),),
                     rep_globs=('*.rst',),
                     renames=(('\.rst$', options.source_suffix),))
        make_link_targets(project_name,
                          options.main_gh_user,
                          options.repo_name,
                          pjoin(out_path, 'gitwash', 'known_projects.inc'),
                          pjoin(out_path, 'gitwash', 'this_project.inc'),
                          options.project_url,
                          options.project_ml_url)
    finally:
        shutil.rmtree(repo_path)


if __name__ == '__main__':
    main()
